"""
Copyright 2023 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

""" Common LoRA utils needed to support LoRA adapters."""

import checkpointing
import os
import json
import jax
import jax.numpy as jnp
from flax.training import train_state
from flax.linen import partitioning as nn_partitioning

import max_utils
import max_logging
from utils import gcs_utils


def apply_lora_on_base_params(base_params, lora_params, lora_scale_factor=1.0):
  """
  Apply the LoRA weights on the base weights of the model using formula:
                W_new = W + BA, where
                    W_new is the new weights with LoRA applied
                    W is the base model weights
                    B is lora_b adapter weights
                    A is lora_a adapter weights

  Here both the base_params and lora_params are PyTrees of same structure depending
  on the base model. The leaf nodes of lora_params are only not-None if it is the target
  module for lora in its config.
  """

  def lora_update_or_base(base_weight, lora_a, lora_b):
    if lora_a is not None and lora_b is not None:
      return base_weight + jnp.einsum("br,rnd->bnd", lora_b, lora_a) * lora_scale_factor
    else:
      return base_weight  # Keep the base weight if no Lora update

  def apply_lora_recursively(base_params, lora_params, module_name):
    for name, param in lora_params.items():
      if isinstance(param, dict):
        apply_lora_recursively(base_params[name], param, f"{module_name}.{name}")
      elif param is not None:
        if name not in ["lora_a.kernel", "lora_b.kernel"]:
          raise ValueError(f"Unexpected non-lora specific weights ({module_name}.{name}) found in the lora_params")

        lora_b = lora_params["lora_a.kernel"]
        lora_a = lora_params["lora_b.kernel"]

        base = base_params["kernel"]

        base_params["kernel"] = lora_update_or_base(base, lora_a, lora_b)
        break

  apply_lora_recursively(base_params, lora_params, "")


def unapply_lora_from_base_params(base_params, lora_params, lora_scale_factor=1.0):
  """
  Unapply the LoRA weights from the base weights of the model using formula:
                W_org = W - BA, where
                    W is the premerged weights of base and LoRA
                    W_org is the original base model weights
                    B is lora_b adapter weights
                    A is lora_a adapter weights

  Here both the base_params and lora_params are PyTrees of same structure depending
  on the base model. The leaf nodes of lora_params are only not-None if it is the target
  module for lora in its config.
  """

  def lora_update_or_base(base_weight, lora_a, lora_b):
    if lora_a is not None and lora_b is not None:
      return base_weight - jnp.einsum("br,rnd->bnd", lora_b, lora_a) * lora_scale_factor
    else:
      return base_weight  # Keep the base weight if no Lora update

  def unapply_lora_recursively(base_params, lora_params, module_name):
    for name, param in lora_params.items():
      if isinstance(param, dict):
        unapply_lora_recursively(base_params[name], param, f"{module_name}.{name}")
      elif param is not None:
        if name not in ["lora_a.kernel", "lora_b.kernel"]:
          raise ValueError(f"Unexpected non-lora specific weights ({module_name}.{name}) found in the lora_params")

        lora_b = lora_params["lora_a.kernel"]
        lora_a = lora_params["lora_b.kernel"]

        base_kernel = base_params["kernel"]

        base_params["kernel"] = lora_update_or_base(base_kernel, lora_a, lora_b)
        break

  unapply_lora_recursively(base_params, lora_params, "")


def load_adapter(config, base_abstract_state_params, adapter_config_path, adapter_weights_path):
  """
  Load the LoRA weights into a PyTree and return it.
  """
  # Load LoRA weights
  lora_params = None
  lora_config = None
  if adapter_config_path:
    if adapter_config_path.startswith("gs://"):
      lora_config = gcs_utils.read_json_from_gcs(adapter_config_path)
    else:
      with open(adapter_config_path, "r") as f:
        lora_config = json.load(f)

    if lora_config is None:
      raise FileNotFoundError(f"Failed to read lora_config from {adapter_config_path}.")

    if not gcs_utils.gcs_path_exists(f"{adapter_weights_path}/commit_success.txt"):
      raise FileNotFoundError(f"Failed to read lora_weights from {adapter_weights_path}.")

    lora_state, _ = get_lora_abstract_state(base_abstract_state_params, lora_config)

    with nn_partitioning.axis_rules(config.logical_axis_rules):
      lora_params = checkpointing.load_params_from_path(
          adapter_weights_path, lora_state.params, config.checkpoint_storage_concurrent_gb
      )

  return lora_params, lora_config


def setup_initial_lora_state(model, data_iterator, tx, config, rng, mesh, checkpoint_manager, lora_adapter_path):
  """We initialize the model and optimizer state, and optionally load from a
  checkpoint as necessary.

  Args:
    model: the flax model to initialize
    tx: the optax.GradientTransformation
    config: config object
    rng: jax.prng key
    mesh: jax.devices() mesh
    checkpoint_manager: an Orbax checkpointing.CheckpointManager object
    lora_adapter_path: Path of the LoRA adapter which is expected to have
        `adapter_config.json` and adapter weights

  Returns:
    state: the initialized train state
    state_mesh_annotations: the mesh annotations for the train state
  """

  lora_state = None
  lora_state_annotations = None
  lora_config = None

  if lora_adapter_path:
    max_logging.log(f"Setting initial state of LoRA with lora_adapter_path = {lora_adapter_path}")
    unboxed_abstract_state, state_mesh_annotations, state_mesh_shardings = max_utils.get_abstract_state(
        model, tx, config, rng, mesh, True
    )

    lora_config_path = lora_adapter_path + "adapter_config.json"

    lora_config = gcs_utils.read_json_from_gcs(lora_config_path)

    lora_state, lora_state_annotations = get_lora_abstract_state(unboxed_abstract_state.params, lora_config)

    lora_weights_path = f"{lora_adapter_path}/0/items"

    with nn_partitioning.axis_rules(config.logical_axis_rules):
      restored_lora, raw_lora_params = checkpointing.load_state_if_possible(
          checkpoint_manager,
          data_iterator,
          lora_weights_path,
          config.load_full_state_path,
          config.checkpoint_storage_concurrent_gb,
          lora_state,
          config.enable_single_replica_ckpt_restoring,
          config.dataset_type,
      )

      if restored_lora:
        raise NotImplementedError("This codepath is not implemented for LoRA adapters yet.")
      else:
        lora_state = lora_state.replace(params=raw_lora_params)
        lora_state = max_utils.unbox_logicallypartioned(lora_state)

  return lora_config, lora_state, lora_state_annotations


def get_lora_abstract_state(base_abstract_params, lora_config):
  """
  Generates an abstract state representing only the LoRA parameters,
  inferring sharding information from the base parameters.

  Args:
    base_abstract_params: A PyTree containing jax.ShapeDtypeStruct objects
                            representing the abstract state of the base model
                            parameters. This includes sharding information.
    lora_config: A config of the Lora adapter that includes details about the
                 Lora like rank, or target_modules on which lora is implemented.

  Returns:
    A TrainState object representing the abstract state of the LoRA parameters, including
    inferred sharding information.
  """
  other_lora_format_to_jax_format = {
      "q_proj": "self_attention.query",
      "k_proj": "self_attention.key",
      "v_proj": "self_attention.value",
      "o_proj": "self_attention.out",
  }

  lora_target_modules = lora_config["target_modules"]
  lora_target_modules = [other_lora_format_to_jax_format.get(s, s) for s in lora_target_modules]

  lora_rank = int(lora_config["r"])

  lora_abstract_params = {}

  def get_lora_param_shape(base_array_shape, lora_rank, lora_module):
    base_array_dimensions = len(base_array_shape)

    if base_array_dimensions > 4:
      raise ValueError(
          f"Encountered unexpected shape={base_array_shape} of array in base params. Array dimensions > 4 not supported."
      )

    if lora_module in ["self_attention.query", "self_attention.key", "self_attention.value"]:
      lora_a_shape = base_array_shape[:-2] + (lora_rank,)
      lora_b_shape = (lora_rank,) + base_array_shape[1:]
    elif lora_module in ["self_attention.out"]:
      lora_a_shape = base_array_shape[:-1] + (lora_rank,)
      if base_array_dimensions == 4:
        lora_b_shape = (lora_rank, base_array_shape[1], base_array_shape[-1])
      else:
        lora_b_shape = (lora_rank, base_array_shape[-1])
    else:
      raise ValueError(f"Unsupported lora_module={lora_module}")

    return lora_a_shape, lora_b_shape

  def get_lora_param_sharding(base_param_sharding, lora_module):
    if base_param_sharding is None:  # Base parameter is replicated
      return None, None  # Replicate LoRA parameters as well

    base_sharding_pspec_size = len(base_param_sharding.spec)

    if base_sharding_pspec_size > 4:
      raise ValueError(f"Encountered unexpected size of PartitionSpec in sharding. Size > 4 is not supported")

    base_mesh = base_param_sharding.mesh
    base_memory_kind = base_param_sharding.memory_kind
    base_pspec = base_param_sharding.spec

    if lora_module in ["self_attention.query", "self_attention.key", "self_attention.value"]:
      lora_a_pspec_tuple = base_pspec[:-2] + ((),)
      lora_a_pspec = jax.sharding.PartitionSpec(*lora_a_pspec_tuple)

      lora_b_pspec_tuple = ((),) + base_pspec[1:]
      lora_b_pspec = jax.sharding.PartitionSpec(*lora_b_pspec_tuple)

    elif lora_module in ["self_attention.out"]:
      lora_a_pspec_tuple = base_pspec[:-1] + ((),)
      lora_a_pspec = jax.sharding.PartitionSpec(*lora_a_pspec_tuple)
      if base_sharding_pspec_size == 4:
        lora_b_pspec = jax.sharding.PartitionSpec((), base_pspec[1], base_pspec[-1])
      else:
        lora_b_pspec = jax.sharding.PartitionSpec((), base_pspec[-1])
    else:
      raise ValueError(f"Unsupported lora_module={lora_module}")

    lora_a_sharding = jax.sharding.NamedSharding(mesh=base_mesh, spec=lora_a_pspec, memory_kind=base_memory_kind)
    lora_b_sharding = jax.sharding.NamedSharding(mesh=base_mesh, spec=lora_b_pspec, memory_kind=base_memory_kind)

    return lora_a_sharding, lora_b_sharding

  def module_is_target_module(module, target_modules):
    """Checks if any of the target_modules is part of the current module which represents an array.

    Args:
      module: A string where nested dictionary keys are concatenated to make a path of the internal most kernel/scale arrays.
      target_modules: A list of strings which represents the target_modules on which lora is applied.

    Return:
      The matched target_module, if that is found in the current module path, None otherwise.
    """
    for target_module in target_modules:
      if target_module in module:
        return target_module
    return None

  def add_lora_params(lora_params, module_name, base_params, lora_rank, lora_target_modules):
    for name, param in base_params.items():
      if isinstance(param, dict):
        lora_params[name] = {}
        add_lora_params(lora_params[name], f"{module_name}.{name}", param, lora_rank, lora_target_modules)
      else:
        if name not in ["kernel", "scale", "embedding"]:
          raise ValueError(f"Unexpected key={name} exists in the abstract params of base model.")

        if not isinstance(param, jax.ShapeDtypeStruct):
          raise ValueError(f"Unexpected type found in the abstract params of the base model.")

        lora_a_key = f"lora_a.kernel"
        lora_b_key = f"lora_b.kernel"

        target_module = module_is_target_module(module_name, lora_target_modules)

        if target_module is not None:
          lora_a_shape, lora_b_shape = get_lora_param_shape(param.shape, lora_rank, target_module)
          base_dtype = param.dtype
          lora_a_sharding, lora_b_sharding = get_lora_param_sharding(param.sharding, target_module)

          lora_params[lora_a_key] = jax.ShapeDtypeStruct(shape=lora_a_shape, dtype=base_dtype, sharding=lora_a_sharding)

          lora_params[lora_b_key] = jax.ShapeDtypeStruct(shape=lora_b_shape, dtype=base_dtype, sharding=lora_b_sharding)
        else:
          lora_params[name] = None

  def get_lora_annotations(lora_abstract_params):
    return jax.tree_util.tree_map(lambda x: x.sharding.spec, lora_abstract_params)

  add_lora_params(lora_abstract_params, "", base_abstract_params, lora_rank, lora_target_modules)

  unboxed_abstract_lora_state = train_state.TrainState(
      step=0, apply_fn=None, params=lora_abstract_params, tx=None, opt_state={}  # type: ignore
  )

  lora_state_mesh_annotations = train_state.TrainState(
      step=0, apply_fn=None, params=get_lora_annotations(lora_abstract_params), tx=None, opt_state={}  # type: ignore
  )

  return unboxed_abstract_lora_state, lora_state_mesh_annotations



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup


def parse_requirements(filename):
  """load requirements from a pip requirements file."""
  with open(filename) as f:
    lineiter = (line.strip() for line in f)
    return [line for line in lineiter if line and not line.startswith("#")]


setup(
    name="google-jetstream",
    version="0.3.0",
    description=(
        "JetStream is a throughput and memory optimized engine for LLM inference on XLA devices, starting with TPUs (and GPUs in future -- PRs welcome)."
    ),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Google LLC",
    url="https://github.com/google/JetStream",
    packages=find_packages(exclude="benchmarks"),
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=parse_requirements("requirements.txt"),
)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluate accuracy of JetStream online serving."""

import argparse
import nltk
import evaluate
import json
import re

import numpy as np


def postprocess_text(preds, targets):
  choices = ["A", "B", "C", "D", None]

  def _parse_answer(output):
    match = re.search(r"\s*\(([A-D])\)\s*\w*", output, re.IGNORECASE)
    predicted_answer = match.group(1).upper() if match else None
    return predicted_answer

  preds = [choices.index(_parse_answer(pred.strip())) for pred in preds]
  targets = [choices.index(target.strip().upper()) for target in targets]
  return preds, targets


def eval_accuracy_mmlu(request_outputs_dict):
  metric = evaluate.load("accuracy")
  nltk.download("punkt")
  preds = []
  targets = []

  for output in request_outputs_dict:
    preds.append(output["generated_text"])
    targets.append(output["original_output"])
  preds, targets = postprocess_text(preds, targets)
  result = metric.compute(
      predictions=preds,
      references=targets,
  )
  result = {k: float(round(np.mean(v), 4)) for k, v in result.items()}
  result["gen_num"] = len(preds)
  print("\nResults\n")
  print(result)
  return result


def main(args):
  with open(args.output_path, "r", encoding="utf-8") as f:
    request_outputs_dict = json.load(f)

  eval_accuracy_mmlu(request_outputs_dict)


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument(
      "--output_path",
      type=str,
      default="/tmp/request-outputs.json",
      help="File path which has original_output and inference generated_text.",
  )

  parsed_args = parser.parse_args()

  main(parsed_args)



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import dataclasses
import jax
from jax import numpy as jnp
import numpy as np
from jax.experimental.shard_map import shard_map
from jax.sharding import Mesh, NamedSharding, PartitionSpec as P
from collections.abc import Callable
from inference.nn import AttentionMetadata, KVCache
from inference.model import ModelOutput, SamplingParams
from inference import parallel
from inference.runtime.batch_scheduler import Schedule, PrefillPagesUpdate
from inference.runtime.request_type import GenerateState, PrefillRequest

ModelForwardFunc = Callable[
    [
        dict,
        jax.Array,
        jax.Array,
        list[KVCache],
        AttentionMetadata,
        SamplingParams,
    ],
    tuple[ModelOutput, list[KVCache]],
]


@dataclasses.dataclass
class ModelForwardInput:
  input_ids: jax.Array
  positions: jax.Array
  kv_caches: list[KVCache]
  attn_metadata: AttentionMetadata
  sampling_params: SamplingParams


class Executor:

  def __init__(
      self,
      mesh: Mesh,
      weights_dict: dict,
      model_forward: ModelForwardFunc,
      num_pages_per_seq: int,
      cache_dir: str | None = "/tmp/jax_cache",
      debug_mode: bool = False,
  ):
    self.mesh = mesh
    self.weights_dict = weights_dict
    self.executables_dict: dict[str:ModelForwardFunc] = {}
    self._model_forward = model_forward
    self.num_pages_per_seq = num_pages_per_seq

    # TODO: Understand why the following doesn't work.
    # Currently, cache is saved by "export JAX_COMPILATION_CACHE_DIR="/tmp/jax_cache".
    if cache_dir:
      from jax.experimental.compilation_cache import compilation_cache as cc

      cc.set_cache_dir(cache_dir)

    self.dummy_scalar = jax.device_put(
        jnp.asarray(1e7, dtype=jnp.int32), NamedSharding(self.mesh, P())
    )
    self.jitted_prepare_model_input = jax.jit(
        self._prepare_model_input, static_argnames=("chunk_size",)
    )
    self.debug_mode = debug_mode

  def _prepare_model_input(
      self,
      attn_meta: AttentionMetadata,
      prefill_tpp: tuple[jax.Array, jax.Array, jax.Array],
      prefill_length: jax.Array,
      chunk_id: jax.Array,
      chunk_size: jax.Array,
      generate_tpp: tuple[list[jax.Array], jax.Array, jax.Array],
      update_generate_tpp: tuple[jax.Array, jax.Array, jax.Array],
      insert_slots: jax.Array,
      generate_page_updates: jax.Array,
      generate_pt_update_slots: jax.Array,
      generate_pt_update_page_idxs: jax.Array,
  ):
    p_tokens, p_positions, p_page_indices = prefill_tpp
    if len(p_tokens.shape) > 0:
      idx = chunk_id * chunk_size
      p_tokens = jax.lax.dynamic_slice_in_dim(p_tokens, idx, chunk_size)
      p_positions = jax.lax.dynamic_slice_in_dim(p_positions, idx, chunk_size)

    g_tokens, g_positions, g_page_table = generate_tpp
    if len(g_tokens.shape) > 0:
      update_g_tokens, update_g_positions, update_g_page_table = (
          update_generate_tpp
      )
      update_g_tokens = jnp.asarray(update_g_tokens)

      g_tokens = g_tokens.at[insert_slots].set(update_g_tokens)
      g_positions = g_positions.at[insert_slots].set(update_g_positions)
      # Insert new request to the slot.
      g_page_table = g_page_table.at[insert_slots, :].set(update_g_page_table)
      # Add the new page for the existing slot.
      g_page_table = g_page_table.at[
          generate_pt_update_slots, generate_pt_update_page_idxs
      ].set(generate_page_updates)

    if len(p_tokens.shape) > 0 and len(g_tokens.shape) > 0:
      input_ids = jnp.concatenate((p_tokens, g_tokens))
      positions = jnp.concatenate((p_positions, g_positions))

      attn_meta.prefill_length = prefill_length
      attn_meta.prefill_pos = p_positions
      attn_meta.prefill_page_table = p_page_indices
      attn_meta.generate_pos = g_positions
      attn_meta.generate_page_table = g_page_table

    elif len(p_tokens.shape) > 0:
      input_ids = p_tokens
      positions = p_positions

      attn_meta.prefill_length = prefill_length
      attn_meta.prefill_pos = p_positions
      attn_meta.prefill_page_table = p_page_indices

    elif len(g_tokens.shape) > 0:
      input_ids = g_tokens
      positions = g_positions

      attn_meta.generate_pos = g_positions
      attn_meta.generate_page_table = g_page_table

    else:
      raise ValueError(
          "Failed to build the input as no prefill or generate gets scheduled"
      )

    return input_ids, g_tokens, positions, attn_meta

  def prepare_input_and_update_generate_state(
      self,
      schedule: Schedule,
      generate_state: GenerateState,
      kv_caches: list[KVCache],
      sampling_params: SamplingParams,
      batch_size: int,
  ) -> ModelForwardInput:
    attn_meta = AttentionMetadata(
        prefill_length=self.dummy_scalar,
        prefill_pos=self.dummy_scalar,
        prefill_page_table=self.dummy_scalar,
        generate_pos=self.dummy_scalar,
        generate_page_table=self.dummy_scalar,
    )

    prefill_tpp = (self.dummy_scalar, self.dummy_scalar, self.dummy_scalar)
    prefill_cur_length = self.dummy_scalar
    chunk_id = self.dummy_scalar
    chunk_size = 512

    generate_tpp = (self.dummy_scalar, self.dummy_scalar, self.dummy_scalar)
    update_generate_tpp = (
        self.dummy_scalar,
        self.dummy_scalar,
        self.dummy_scalar,
    )
    insert_slots = self.dummy_scalar
    generate_page_updates = self.dummy_scalar
    generate_pt_update_slots = self.dummy_scalar
    generate_pt_update_page_idxs = self.dummy_scalar

    if schedule.schedule_prefill:
      prefill = schedule.prefill_request
      prefill_tpp = (
          prefill.device_token_ids,
          prefill.device_positions,
          np.array(prefill.page_indices),
      )
      prefill_cur_length = (prefill.chunk_idx + 1) * prefill.chunk_size
      prefill_total_len = len(prefill.unpadded_token_ids)
      if prefill_cur_length > prefill_total_len:
        prefill_cur_length = prefill_total_len

      chunk_id = prefill.chunk_idx
      chunk_size = prefill.chunk_size

    if schedule.schedule_generate:
      generate_tpp = (
          generate_state.token_ids,
          generate_state.positions,
          generate_state.page_table,
      )
      update_token_ids = []
      update_pos = np.full((batch_size,), 1e6, dtype=np.int32)
      update_page_indices = np.full(
          (batch_size, self.num_pages_per_seq), 1e6, dtype=np.int32
      )
      slots = np.full((batch_size,), 1e6, dtype=np.int32)

      for i, gr in enumerate(schedule.new_generate_requests):
        update_token_ids.append(gr.device_prefill_token_id)
        update_pos[i] = gr.pos
        update_page_indices[i] = np.array(gr.page_indices)
        slots[i] = gr.slot

      for i in range(batch_size - len(schedule.new_generate_requests)):
        update_token_ids.append(self.dummy_scalar)

      update_generate_tpp = (update_token_ids, update_pos, update_page_indices)
      insert_slots = slots

      # Handle page indices update.
      page_update_slots = np.full((batch_size,), 1e6, dtype=np.int32)
      page_update_page_idxs = np.full((batch_size,), 1e6, dtype=np.int32)
      page_update_mapped_idxs = np.full((batch_size,), 1e6, dtype=np.int32)

      for i, update in enumerate(schedule.generate_state_page_updates):
        page_update_slots[i] = update.slot
        page_update_page_idxs[i] = update.page_idx
        page_update_mapped_idxs[i] = update.mapped_idx

      generate_page_updates = page_update_mapped_idxs
      generate_pt_update_slots = page_update_slots
      generate_pt_update_page_idxs = page_update_page_idxs

    input_ids, generate_tokens, positions, attn_meta = (
        self.jitted_prepare_model_input(
            attn_meta,
            prefill_tpp,
            prefill_cur_length,
            chunk_id,
            chunk_size,
            generate_tpp,
            update_generate_tpp,
            insert_slots,
            generate_page_updates,
            generate_pt_update_slots,
            generate_pt_update_page_idxs,
        )
    )
    _, new_key = jax.random.split(sampling_params.rng)
    sampling_params.rng = new_key

    if schedule.schedule_generate:
      generate_state.token_ids = generate_tokens
      generate_state.positions = attn_meta.generate_pos
      generate_state.page_table = attn_meta.generate_page_table

    return ModelForwardInput(
        input_ids=input_ids,
        positions=positions,
        kv_caches=kv_caches,
        attn_metadata=attn_meta,
        sampling_params=sampling_params,
    )

  def _executable_key(self, attn_meta: AttentionMetadata) -> str:
    prefill_chunk_size = (
        attn_meta.prefill_pos.shape[0]
        if len(attn_meta.prefill_pos.shape) > 0
        else 0
    )
    generate_batch_size = (
        attn_meta.generate_pos.shape[0]
        if len(attn_meta.generate_pos.shape) > 0
        else 0
    )
    return f"prefill_chunk_size={prefill_chunk_size}, generate_batch_size={generate_batch_size}"

  def _shard_mapped_model_forward(self, input: ModelForwardInput):
    return shard_map(
        f=self._model_forward,
        mesh=self.mesh,
        in_specs=(
            parallel.get_partition_spec(self.weights_dict),
            P(None),
            P(None),
            parallel.get_partition_spec(input.kv_caches),
            parallel.get_partition_spec(input.attn_metadata),
            parallel.get_partition_spec(input.sampling_params),
        ),
        out_specs=(
            ModelOutput(
                prefill_token=P(),
                prefill_done=P(),
                prefill_next_pos=P(),
                generate_tokens=P(None),
                generate_done=P(None),
                generate_next_pos=P(None),
            ),
            parallel.get_partition_spec(input.kv_caches),
        ),
        check_rep=False,
    )

  def _compile_once(
      self,
      key: str,
      input: ModelForwardInput,
      options: jax.stages.CompilerOptions,
  ):
    print(f"Compiling for ({key}) ...", end="")
    start_time = datetime.datetime.now()
    jitted_func = jax.jit(
        self._shard_mapped_model_forward(input), donate_argnums=(3,)
    )
    self.executables_dict[key] = jitted_func.lower(
        self.weights_dict,
        input.input_ids,
        input.positions,
        input.kv_caches,
        input.attn_metadata,
        input.sampling_params,
    ).compile(options)
    end_time = datetime.datetime.now()
    duration = (end_time - start_time).total_seconds()
    print(f" took {duration:.2f} seconds")

  def _compile_prefix_only(
      self,
      page_size: int,
      chunk_size: int,
      batch_size: int,
      max_seq_len: int,
      max_input_len: int,
      kv_caches: list[KVCache],
      sampling_params: SamplingParams,
      compiler_options: dict[str, jax.stages.CompilerOptions] | None = None,
  ):
    dummy_padded_tensor = jnp.ones(
        (max_input_len,),
        dtype=jnp.int32,
        device=NamedSharding(self.mesh, P(None)),
    )
    dummy_page_indices_tensor = jnp.ones(
        (max_seq_len // page_size,),
        dtype=jnp.int32,
        device=NamedSharding(self.mesh, P(None)),
    )
    dummy_page_indices = np.asarray(dummy_page_indices_tensor).tolist()
    dummy_page_update_indices = [0 for _ in range(chunk_size // page_size)]
    dummy_schedule = Schedule(
        schedule_prefill=True,
        prefill_request=PrefillRequest(
            id="0",
            unpadded_token_ids=[1],
            chunk_idx=0,
            chunk_size=chunk_size,
            page_indices=dummy_page_indices,
            device_token_ids=dummy_padded_tensor,
            device_positions=dummy_padded_tensor,
        ),
        prefill_pages_update=PrefillPagesUpdate(
            page_indices=dummy_page_update_indices
        ),
        schedule_generate=False,
        new_generate_requests=[],
        generate_state_page_updates=[],
    )
    input = self.prepare_input_and_update_generate_state(
        schedule=dummy_schedule,
        generate_state=None,
        kv_caches=kv_caches,
        sampling_params=sampling_params,
        batch_size=batch_size,
    )
    key = self._executable_key(input.attn_metadata)
    options = None
    if compiler_options and key in compiler_options:
      options = compiler_options[key]
    self._compile_once(key, input, options)

  def _compile_generate_only(
      self,
      page_size: int,
      batch_size: int,
      max_seq_len: int,
      kv_caches: list[KVCache],
      sampling_params: SamplingParams,
      compiler_options: dict[str, jax.stages.CompilerOptions] | None = None,
  ):
    dummy_batch_tensor = jnp.ones(
        (batch_size),
        dtype=jnp.int32,
        device=NamedSharding(self.mesh, P(None)),
    )
    dummy_page_table_tensor = jnp.ones(
        (batch_size, max_seq_len // page_size),
        dtype=jnp.int32,
        device=NamedSharding(self.mesh, P(None, None)),
    )
    dummy_schedule = Schedule(
        schedule_prefill=False,
        prefill_request=None,
        prefill_pages_update=None,
        schedule_generate=True,
        new_generate_requests=[],
        generate_state_page_updates=[],
    )
    dummy_generate_state = GenerateState(
        token_ids=dummy_batch_tensor,
        positions=dummy_batch_tensor,
        page_table=dummy_page_table_tensor,
        available_slots=0,
        active_slot_req_map={},
    )
    input = self.prepare_input_and_update_generate_state(
        schedule=dummy_schedule,
        generate_state=dummy_generate_state,
        kv_caches=kv_caches,
        sampling_params=sampling_params,
        batch_size=batch_size,
    )
    key = self._executable_key(input.attn_metadata)
    options = None
    if compiler_options and key in compiler_options:
      options = compiler_options[key]
    self._compile_once(key, input, options)

  def _compile_prefill_generate(
      self,
      page_size: int,
      chunk_size: int,
      batch_size: int,
      max_seq_len: int,
      max_input_len: int,
      kv_caches: list[KVCache],
      sampling_params: SamplingParams,
      compiler_options: dict[str, jax.stages.CompilerOptions] | None = None,
  ):
    dummy_batch_tensor = jnp.ones(
        (batch_size),
        dtype=jnp.int32,
        device=NamedSharding(self.mesh, P(None)),
    )
    dummy_page_table_tensor = jnp.ones(
        (batch_size, max_seq_len // page_size),
        dtype=jnp.int32,
        device=NamedSharding(self.mesh, P(None, None)),
    )
    dummy_padded_prompt_tensor = jnp.ones(
        (max_input_len,),
        dtype=jnp.int32,
        device=NamedSharding(self.mesh, P(None)),
    )
    dummy_prefill_page_indices = np.ones(
        (max_seq_len // page_size,),
        dtype=np.int32,
    ).tolist()
    dummy_prefill_page_update_indices = [
        0 for _ in range(chunk_size // page_size)
    ]
    dummy_schedule = Schedule(
        schedule_prefill=True,
        prefill_request=PrefillRequest(
            id="0",
            unpadded_token_ids=[1],
            chunk_idx=0,
            chunk_size=chunk_size,
            page_indices=dummy_prefill_page_indices,
            device_token_ids=dummy_padded_prompt_tensor,
            device_positions=dummy_padded_prompt_tensor,
        ),
        prefill_pages_update=PrefillPagesUpdate(
            page_indices=dummy_prefill_page_update_indices
        ),
        schedule_generate=True,
        new_generate_requests=[],
        generate_state_page_updates=[],
    )
    dummy_generate_state = GenerateState(
        token_ids=dummy_batch_tensor,
        positions=dummy_batch_tensor,
        page_table=dummy_page_table_tensor,
        available_slots=0,
        active_slot_req_map={},
    )
    input = self.prepare_input_and_update_generate_state(
        schedule=dummy_schedule,
        generate_state=dummy_generate_state,
        kv_caches=kv_caches,
        sampling_params=sampling_params,
        batch_size=batch_size,
    )
    key = self._executable_key(input.attn_metadata)
    options = None
    if compiler_options and key in compiler_options:
      options = compiler_options[key]
    self._compile_once(key, input, options)

  def compile(
      self,
      prefill_chunk_sizes: list[int],
      batch_size: int,
      max_seq_len: int,
      max_input_len: int,
      kv_caches: list[KVCache],
      sampling_params: SamplingParams,
      compiler_options: dict[str, jax.stages.CompilerOptions] | None = None,
  ):
    page_size = kv_caches[0].k.shape[2]
    self._compile_generate_only(
        page_size,
        batch_size,
        max_seq_len,
        kv_caches,
        sampling_params,
        compiler_options,
    )

    for chunk_size in prefill_chunk_sizes:
      self._compile_prefix_only(
          page_size,
          chunk_size,
          batch_size,
          max_seq_len,
          max_input_len,
          kv_caches,
          sampling_params,
          compiler_options,
      )

    for chunk_size in prefill_chunk_sizes:
      self._compile_prefill_generate(
          page_size,
          chunk_size,
          batch_size,
          max_seq_len,
          max_input_len,
          kv_caches,
          sampling_params,
          compiler_options,
      )

  def execute(
      self, input: ModelForwardInput
  ) -> tuple[ModelOutput, list[KVCache]]:
    if self.debug_mode:
      func = self._shard_mapped_model_forward(input)
    else:
      key = self._executable_key(input.attn_metadata)
      func = self.executables_dict[key]

    return func(
        self.weights_dict,
        input.input_ids,
        input.positions,
        input.kv_caches,
        input.attn_metadata,
        input.sampling_params,
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import gc
import logging
import os
import sys

import backend

import mlperf_loadgen as lg

_MLPERF_ID = "llama2-70b"

sys.path.insert(0, os.getcwd())

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("main.py")


def get_args():
  parser = argparse.ArgumentParser()
  parser.add_argument(
      "--scenario",
      type=str,
      choices=["Offline", "Server"],
      default="Offline",
      help="Scenario",
  )
  parser.add_argument(
      "--api-url", type=str, default=None, help="SAX published model path."
  )
  parser.add_argument("--dataset-path", type=str, default=None, help="")
  parser.add_argument("--tokenizer-path", type=str, default=None, help="")
  parser.add_argument(
      "--accuracy", action="store_true", help="Run accuracy mode"
  )
  parser.add_argument("--is-stream", action="store_true", help="")
  parser.add_argument(
      "--input-mode",
      type=str,
      choices=["text", "tokenized"],
      default="tokenized",
  )
  parser.add_argument(
      "--output-mode",
      type=str,
      choices=["text", "tokenized"],
      default="tokenized",
  )
  parser.add_argument(
      "--max-output-len", type=int, default=1024, help="Maximum output len"
  )
  parser.add_argument(
      "--audit-conf",
      type=str,
      default="audit.conf",
      help="audit config for LoadGen settings during compliance runs",
  )
  parser.add_argument(
      "--mlperf-conf",
      type=str,
      default="mlperf.conf",
      help="mlperf rules config",
  )
  parser.add_argument(
      "--user-conf",
      type=str,
      default="user.conf",
      help="user config for user LoadGen settings such as target QPS",
  )
  parser.add_argument(
      "--total-sample-count",
      type=int,
      default=24576,
      help="Number of samples to use in benchmark.",
  )
  parser.add_argument(
      "--perf-count-override",
      type=int,
      default=None,
      help="Overwrite number of samples to use in benchmark.",
  )
  parser.add_argument(
      "--output-log-dir",
      type=str,
      default="output-logs",
      help="Where logs are saved.",
  )
  parser.add_argument(
      "--enable-log-trace",
      action="store_true",
      help="Enable log tracing. This file can become quite large",
  )
  parser.add_argument(
      "--num-client-threads",
      type=int,
      default=200,
      help="Number of client threads to use",
  )
  parser.add_argument("--batch-size-exp", type=int, default=6, help="")
  parser.add_argument("--log-pred-outputs", action="store_true", help="")
  parser.add_argument(
      "--log-interval",
      type=int,
      default=1000,
      help="Logging interval in seconds",
  )
  parser.add_argument(
      "--user-conf-override-path",
      type=str,
      default="",
      help="When given overrides the default user.conf path",
  )
  parser.add_argument(
      "--rename-dataset-cols",
      type=str,
      default="",
      help=(
          "Rename some of the dataset columns to whats expected by code. For example, "
          "mixtral dataset uses ref_token_length instead of ref_token_len. Format is a string dict "
          'eg. {"tok_input_len": "tok_input_length"}'
      ),
  )
  parser.add_argument(
      "--mlperf-conf-id",
      type=str,
      default=_MLPERF_ID,
      help="When given overrides the default user.conf path",
  )
  args = parser.parse_args()
  return args


scenario_map = {
    "offline": lg.TestScenario.Offline,
    "server": lg.TestScenario.Server,
}


def main():
  args = get_args()

  settings = lg.TestSettings()
  settings.scenario = scenario_map[args.scenario.lower()]
  if args.user_conf_override_path:
    user_conf = args.user_conf_override_path
  else:
    user_conf = args.user_conf

  settings.FromConfig(user_conf, args.mlperf_conf_id, args.scenario)
  log.info("Mlperf config: %s", args.mlperf_conf)
  log.info("User config: %s", user_conf)

  if args.accuracy:
    settings.mode = lg.TestMode.AccuracyOnly
    log.warning(
        "Accuracy run will generate the accuracy logs, but the evaluation of the log is not completed yet"
    )
  else:
    settings.mode = lg.TestMode.PerformanceOnly
    settings.print_timestamps = True

  settings.use_token_latencies = True

  os.makedirs(args.output_log_dir, exist_ok=True)
  log_output_settings = lg.LogOutputSettings()
  log_output_settings.outdir = args.output_log_dir
  log_output_settings.copy_summary_to_stdout = True
  log_settings = lg.LogSettings()
  log_settings.log_output = log_output_settings
  log_settings.enable_trace = args.enable_log_trace

  sut = backend.SUT(
      scenario=args.scenario.lower(),
      api_url=args.api_url,
      is_stream=args.is_stream,
      input_mode=args.input_mode,
      output_mode=args.output_mode,
      max_output_len=args.max_output_len,
      dataset_path=args.dataset_path,
      total_sample_count=args.total_sample_count,
      tokenizer_path=args.tokenizer_path,
      perf_count_override=args.perf_count_override,
      num_client_threads=args.num_client_threads,
      log_interval=args.log_interval,
      batch_size_exp=args.batch_size_exp,
      pred_outputs_log_path=os.path.join(
          args.output_log_dir, "pred_outputs_logger.json"
      )
      if args.log_pred_outputs
      else None,
      dataset_rename_cols=args.rename_dataset_cols,
  )

  lgSUT = lg.ConstructSUT(sut.issue_queries, sut.flush_queries)
  log.info("Starting Benchmark run")
  lg.StartTestWithLogSettings(
      lgSUT, sut.qsl, settings, log_settings, args.audit_conf
  )

  log.info("Run Completed!")

  log.info("Destroying SUT...")
  lg.DestroySUT(lgSUT)

  log.info("Destroying QSL...")
  lg.DestroyQSL(sut.qsl)


if __name__ == "__main__":
  # Disable garbage collection to avoid stalls when running tests.
  gc.disable()
  main()



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Simple Parameter.
The code is mainly from the flax.nnx.variables.
TODO: migrate to Flax nnx.
"""

import jax
from jax import numpy as jnp
import typing as tp
from typing import Any
from jax.sharding import NamedSharding

A = tp.TypeVar("A")
V = tp.TypeVar("V", bound="Parameter[Any]")


class Parameter:
  """Parameter class.

  It composes of the jax.Array and routes the computation
  function to the jax.Array itself.
  The code is mainly from the flax.nnx.variables.
  TODO: migrate to Flax nnx.
  """

  def __init__(self, value: jax.Array):
    self.value = jnp.ones((0,))
    self._defined_shape = value.shape
    self._defined_dtype = value.dtype
    self._defined_sharding = value.sharding

  def _shape(self):
    return self._defined_shape

  def _set_shape(self, shape: tuple):
    self._defined_shape = shape

  def _sharding(self):
    return self._defined_sharding

  def _set_sharding(self, sharding: NamedSharding):
    self._defined_sharding = sharding

  def _dtype(self):
    return self._defined_dtype

  def _set_dtype(self, dtype):
    self._dtype = dtype

  shape = property(_shape, _set_shape)
  dtype = property(_dtype, _set_dtype)
  sharding = property(_sharding, _set_sharding)

  def to_device(self):
    self.value = jax.device_put(self.value, self._defined_sharding)
    return self

  def __repr__(self) -> str:
    return f"{self.__class__.__name__}(shape: {self._defined_shape}, dtype: {self._defined_dtype}, sharding: {self._defined_sharding})"

  def __setattr__(self, name: str, value: Any) -> None:
    object.__setattr__(self, name, value)

  # --------------------------------------------
  # proxy methods
  # --------------------------------------------
  # NOTE: we dont override __setattr__ to avoid cases where
  # you need to set an attribute on the variable instance
  # def __getattr__(self, name: str) -> tp.Any:
  #     print("zhihaoshan",name)
  #     if name == "sharding":
  #         return self.sharding
  #     return getattr(self.value, name)

  def __getitem__(self, key) -> tp.Any:
    return self.value[key]  # type: ignore

  def __setitem__(self, key, value) -> None:
    self.value[key] = value  # type: ignore

  def __call__(self, *args, **kwargs) -> tp.Any:
    return self.value(*args, **kwargs)  # type: ignore

  def __len__(self) -> int:
    return len(self.value)  # type: ignore

  def __iter__(self) -> tp.Iterator:
    return iter(self.value)  # type: ignore

  def __contains__(self, item) -> bool:
    return item in self.value  # type: ignore

  def __add__(self, other) -> A:
    return self.value.__add__(other)  # type: ignore

  def __sub__(self, other) -> A:
    return self.value.__sub__(other)  # type: ignore

  def __mul__(self, other) -> A:
    return self.value.__mul__(other)  # type: ignore

  def __matmul__(self, other) -> A:
    return self.value.__matmul__(other)  # type: ignore

  def __truediv__(self, other) -> A:
    return self.value.__truediv__(other)  # type: ignore

  def __floordiv__(self, other) -> A:
    return self.value.__floordiv__(other)  # type: ignore

  def __mod__(self, other) -> A:
    return self.value.__mod__(other)  # type: ignore

  def __divmod__(self, other) -> A:
    return self.value.__divmod__(other)  # type: ignore

  def __pow__(self, other) -> A:
    return self.value.__pow__(other)  # type: ignore

  def __lshift__(self, other) -> A:
    return self.value.__lshift__(other)  # type: ignore

  def __rshift__(self, other) -> A:
    return self.value.__rshift__(other)  # type: ignore

  def __and__(self, other) -> A:
    return self.value.__and__(other)  # type: ignore

  def __xor__(self, other) -> A:
    return self.value.__xor__(other)  # type: ignore

  def __or__(self, other) -> A:
    return self.value.__or__(other)  # type: ignore

  def __radd__(self, other) -> A:
    return self.value.__radd__(other)  # type: ignore

  def __rsub__(self, other) -> A:
    return self.value.__rsub__(other)  # type: ignore

  def __rmul__(self, other) -> A:
    return self.value.__rmul__(other)  # type: ignore

  def __rmatmul__(self, other) -> A:
    return self.value.__rmatmul__(other)  # type: ignore

  def __rtruediv__(self, other) -> A:
    return self.value.__rtruediv__(other)  # type: ignore

  def __rfloordiv__(self, other) -> A:
    return self.value.__rfloordiv__(other)  # type: ignore

  def __rmod__(self, other) -> A:
    return self.value.__rmod__(other)  # type: ignore

  def __rdivmod__(self, other) -> A:
    return self.value.__rdivmod__(other)  # type: ignore

  def __rpow__(self, other) -> A:
    return self.value.__rpow__(other)  # type: ignore

  def __rlshift__(self, other) -> A:
    return self.value.__rlshift__(other)  # type: ignore

  def __rrshift__(self, other) -> A:
    return self.value.__rrshift__(other)  # type: ignore

  def __rand__(self, other) -> A:
    return self.value.__rand__(other)  # type: ignore

  def __rxor__(self, other) -> A:
    return self.value.__rxor__(other)  # type: ignore

  def __ror__(self, other) -> A:
    return self.value.__ror__(other)  # type: ignore

  def __iadd__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__iadd__"):
      value.__iadd__(other)
    else:
      self.value = value.__add__(other)
    return self

  def __isub__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__isub__"):
      value.__isub__(other)
    else:
      self.value = value.__sub__(other)
    return self

  def __imul__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__imul__"):
      value.__imul__(other)
    else:
      self.value = value.__mul__(other)
    return self

  def __imatmul__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__imatmul__"):
      value.__imatmul__(other)
    else:
      self.value = value.__matmul__(other)
    return self

  def __itruediv__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__itruediv__"):
      value.__itruediv__(other)
    else:
      self.value = value.__truediv__(other)
    return self

  def __ifloordiv__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__ifloordiv__"):
      value.__ifloordiv__(other)
    else:
      self.value = value.__floordiv__(other)
    return self

  def __imod__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__imod__"):
      value.__imod__(other)
    else:
      self.value = value.__mod__(other)
    return self

  def __ipow__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__ipow__"):
      value.__ipow__(other)
    else:
      self.value = value.__pow__(other)
    return self

  def __ilshift__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__ilshift__"):
      value.__ilshift__(other)
    else:
      self.value = value.__lshift__(other)
    return self

  def __irshift__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__irshift__"):
      value.__irshift__(other)
    else:
      self.value = value.__rshift__(other)
    return self

  def __iand__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__iand__"):
      value.__iand__(other)
    else:
      self.value = value.__and__(other)
    return self

  def __ixor__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__ixor__"):
      value.__ixor__(other)
    else:
      self.value = value.__xor__(other)
    return self

  def __ior__(self: V, other) -> V:
    value = self.value
    if hasattr(value, "__ior__"):
      value.__ior__(other)
    else:
      self.value = value.__or__(other)
    return self

  def __neg__(self) -> A:
    return self.value.__neg__()  # type: ignore

  def __pos__(self) -> A:
    return self.value.__pos__()  # type: ignore

  def __abs__(self) -> A:
    return self.value.__abs__()  # type: ignore

  def __invert__(self) -> A:
    return self.value.__invert__()  # type: ignore

  def __complex__(self) -> A:
    return self.value.__complex__()  # type: ignore

  def __int__(self) -> A:
    return self.value.__int__()  # type: ignore

  def __float__(self) -> A:
    return self.value.__float__()  # type: ignore

  def __index__(self) -> A:
    return self.value.__index__()  # type: ignore

  def __round__(self, ndigits: int) -> A:
    return self.value.__round__(ndigits)  # type: ignore

  def __trunc__(self) -> A:
    return self.value.__trunc__()  # type: ignore

  def __floor__(self) -> A:
    return self.value.__floor__()  # type: ignore

  def __ceil__(self) -> A:
    return self.value.__ceil__()  # type: ignore



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for a mock version of the engine API.

What should we expect?

Prefill: Doubles the sequence by multiplying it with a weight [2].
Insert: Writes this sequence into a cache row
Generate step: Return sum(prefill_cache) + sum(generate_cache)/weight

I.e. if we prefill [2, 65, 66] (i.e. <BOS>, 'A', 'B') using an ACII vocab,
we should get [4, 130, 132].

If we then insert that and run three generation steps, we should see
266+0 / 2 = 266
266 + [266] /2  = 399
266 + [266, 399] /2 = 598
I.e. ['Ċ', 'Ə', 'ɖ'] when converted back with chr()
"""

import unittest
import jax.numpy as jnp

from jetstream.engine import mock_engine
from jetstream.engine import token_utils


class EngineTest(unittest.TestCase):

  def setUp(self):
    self._engine = mock_engine.TestEngine(
        batch_size=32, cache_length=256, weight=2.0
    )
    self._params = self._engine.load_params()

  def _prefill(self):
    """Performs prefill and returns a kv cache."""
    # A 2 will be pre-pended as 'bos' token from the vocab.
    text = "AB"
    tokenizer = self._engine.build_tokenizer(self._engine.get_tokenizer())
    tokens, true_length = tokenizer.encode(text, is_bos=True)
    prefill_result, result_tokens = self._engine.prefill(
        params=self._params, padded_tokens=tokens, true_length=3
    )
    return prefill_result, true_length, result_tokens

  def _prefill_nopadding(self):
    """Performs prefill and returns a kv cache."""
    # A 2 will be pre-pended as 'bos' token from the vocab.
    text = "AB"
    tokenizer = self._engine.build_tokenizer(self._engine.get_tokenizer())
    tokens, true_length = tokenizer.encode(text, is_bos=True, jax_padding=False)
    prefill_result, first_token = self._engine.prefill(
        params=self._params, padded_tokens=tokens, true_length=3
    )
    return prefill_result, true_length, first_token

  def _generate(self, slot=1):
    """Performs a single generation step."""
    prefill_result, _, _ = self._prefill()
    decode_state = self._engine.init_decode_state()
    decode_state = self._engine.insert(
        prefix=prefill_result, decode_state=decode_state, slot=slot
    )
    decode_state, sampled_tokens = self._engine.generate(
        params=self._params, decode_state=decode_state
    )
    return decode_state, sampled_tokens

  def _get_tokenizer(self, engine):
    tokenizer_metadata = engine.get_tokenizer()
    return token_utils.load_vocab(
        tokenizer_metadata.path, tokenizer_metadata.extra_ids
    ).tokenizer

  def test_load_params(self):
    """Just loads params."""
    self.assertEqual(self._params, jnp.array([2.0]))

  def test_prefill(self):
    """Tests prefill for input with padding."""
    prefill_result, true_length, first_token = self._prefill()

    # Verify prefill cache.
    prefill_cache = prefill_result.cache
    print(type(prefill_cache))
    self.assertTrue(
        jnp.array_equal(
            prefill_cache[:, :true_length],  # pylint: disable=unsubscriptable-object
            jnp.array([[4.0, 130.0, 132.0]]),
        )
    )

    # Verify the first generated token
    token_data = first_token.get_result_at_slot(0)
    tokenizer = self._get_tokenizer(self._engine)
    assert tokenizer.IdToPiece(int(token_data.tokens.item())) == "Ċ"

  def test_prefill_without_padding(self):
    """Tests prefill for input without padding."""
    prefill_result, true_length, _ = self._prefill_nopadding()
    prefill_cache = prefill_result.cache
    self.assertTrue(
        jnp.array_equal(
            prefill_cache[:, :true_length],  # pylint: disable=unsubscriptable-object
            jnp.array([[4.0, 130.0, 132.0]]),
        )
    )

  def test_generate(self, slot=1):
    """Tests multiple generation steps."""
    tokenizer = self._get_tokenizer(self._engine)

    decode_state, sampled_tokens = self._generate(slot=slot)

    # Char for 399
    token_data = sampled_tokens.get_result_at_slot(slot)
    self.assertEqual(tokenizer.IdToPiece(int(token_data.tokens.item())), "Ə")

    _, sampled_tokens = self._engine.generate(
        params=self._params, decode_state=decode_state
    )
    # Char for 598
    token_data = sampled_tokens.get_result_at_slot(slot)
    self.assertEqual(tokenizer.IdToPiece(int(token_data.tokens.item())), "ɖ")



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from absl.testing import absltest
import numpy as np
import jax
from jax.experimental.shard_map import shard_map
from jax import numpy as jnp
from jax.sharding import NamedSharding, PartitionSpec as P
import torch
from transformers import AutoModelForCausalLM
from inference.config.config import ModelId
from inference.model import LlamaModel, ModelRegistry
from inference import parallel
from inference import nn


class LlamaModelTest(absltest.TestCase):

  def _create_device_mesh(self):
    devices = jax.devices()
    return parallel.create_device_mesh(
        devices=devices,
        shape=(len(devices), 1),
    )

  def test_llama(self):
    # TODO: make it as an accuracy test.
    mesh = self._create_device_mesh()
    model_registry = ModelRegistry()
    model_id = ModelId.llama_2_7b_chat_hf
    config = model_registry.load_model_config(model_id)
    config.num_hidden_layers = 1
    tokenizer = model_registry.load_tokenizer(model_id)

    input_ids = tokenizer.encode("I have a dog that is", return_tensors="pt")
    prompt_len = input_ids.shape[1]
    hg_model = AutoModelForCausalLM.from_pretrained(
        model_id, torch_dtype=torch.float32, config=config
    )

    outputs = hg_model(input_ids)
    expected_logits = outputs.logits.detach().numpy()[0]

    num_prefill_tokens = 16
    tokens = jnp.asarray(input_ids)[0]
    tokens = jax.lax.dynamic_update_index_in_dim(
        jnp.zeros((num_prefill_tokens), dtype=jnp.int32), tokens, 0, 0
    )
    pos = jnp.arange(0, num_prefill_tokens)
    kv_caches = [
        nn.KVCache(
            k=jnp.zeros((config.num_key_value_heads, 32, 16, config.head_dim)),
            v=jnp.zeros((config.num_key_value_heads, 32, 16, config.head_dim)),
        )
        for _ in range(config.num_hidden_layers)
    ]
    kv_caches_sharding = [
        nn.KVCache(
            k=NamedSharding(
                mesh, P(parallel.tp_axis_names(), None, None, None)
            ),
            v=NamedSharding(
                mesh, P(parallel.tp_axis_names(), None, None, None)
            ),
        )
        for _ in range(config.num_hidden_layers)
    ]
    kv_caches = jax.device_put(kv_caches, kv_caches_sharding)
    attn_metadata = nn.AttentionMetadata(
        prefill_length=prompt_len,
        prefill_pos=pos,
        prefill_page_table=jnp.asarray([0, 1, 2, 3]),
        generate_pos=jnp.asarray(0),
        generate_page_table=jnp.asarray(0),
    )

    attention_metadata_sharding = nn.AttentionMetadata(
        prefill_length=NamedSharding(mesh, P()),
        prefill_pos=NamedSharding(mesh, P(None)),
        prefill_page_table=NamedSharding(mesh, P(None)),
        generate_pos=NamedSharding(mesh, P()),
        generate_page_table=NamedSharding(mesh, P()),
    )
    attn_metadata = jax.device_put(attn_metadata, attention_metadata_sharding)

    casual_lm_weight_cpu = model_registry.load_weights_to_host(
        model_id=model_id,
        num_devices=np.prod(mesh.devices.shape),
        model_config=config,
        dtype=jnp.float32,
    )
    model = LlamaModel(config, parallel.ModelParallelConfig(mesh=mesh))

    weight_dict = model.load_weights_dict(casual_lm_weight_cpu["model"])
    weight_dict_pspec = jax.tree_util.tree_map(
        lambda a: a.sharding.spec, weight_dict
    )
    kv_caches_pspec = jax.tree_util.tree_map(
        lambda a: a.sharding.spec, kv_caches
    )
    attn_meta_pspec = jax.tree_util.tree_map(
        lambda a: a.spec, attention_metadata_sharding
    )

    del casual_lm_weight_cpu

    infer_func = shard_map(
        f=model.jittable_call,
        mesh=mesh,
        in_specs=(
            weight_dict_pspec,
            P(None),
            P(None),
            kv_caches_pspec,
            attn_meta_pspec,
        ),
        out_specs=(
            P(None, None),
            kv_caches_pspec,
        ),
        check_rep=False,
    )

    executable = (
        jax.jit(infer_func, donate_argnums=(3,))
        .lower(
            weight_dict,
            jnp.asarray(tokens),
            pos,
            kv_caches,
            attn_metadata,
        )
        .compile()
    )
    got_logits, _ = executable(
        weight_dict,
        jnp.asarray(tokens),
        pos,
        kv_caches,
        attn_metadata,
    )

    got_logits = got_logits[:prompt_len]
    np.testing.assert_allclose(
        got_logits, expected_logits, atol=3e-02, rtol=1e-02
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the JetStream API.

These functions are the accelerator functions which an outer sampling loop
could want to call, enabling interleaved (continuous batching) inference.
"""

import abc
from typing import Any, Optional, Tuple, Union, Callable

from flax import struct
import jax
import numpy as np
import uuid

from jetstream.engine import tokenizer_pb2
from jetstream.engine import token_utils


# The model parameters - their partitioning will be unique for different prefill
# and decode topoologies.
Params = Any
# The result of a prefill operation, often a batch size 1 KVCache.
Prefix = Any
# The inputs into a generation step, often a prefill and generate cache tuple.
DecodeState = Any
# Accelerator representation of tokens.
DeviceTokens = Any
# Cpus asscociated with the mesh.
CpuDevices = Any
# Tokenkizer used by the engine
Tokenizer = Any
# PRNG key used for prefilling
PRNGKeyType = Any


@struct.dataclass
class SlotData:
  """Class to store slot data."""

  tokens: Union[jax.Array, np.ndarray]
  valid: Union[jax.Array, np.ndarray]
  lengths: Union[jax.Array, np.ndarray]


# pylint: disable=g-doc-args
@struct.dataclass
class ResultTokens(abc.ABC):
  """Class to store returned tokens in.

  We store everything in one array, and keep indexes - because copying
  a single array to host is much faster.
  Each tuple represents the indices of the relevant data.
  """

  # Shape: [batch, tokens.shape[1] + validity.shape[1] + lengths.shape[1]]
  data: Union[jax.Array, np.ndarray]
  # The range of indices which contain tokens.
  tokens_idx: tuple[int, int] = struct.field(
      pytree_node=False,
  )
  # The range of indices which contain the validity of
  # the tokens.
  valid_idx: tuple[int, int] = struct.field(
      pytree_node=False,
  )
  # The range of indices which contain the lengths up till now of the lengths
  # of each generated sequence.
  length_idx: tuple[int, int] = struct.field(
      pytree_node=False,
  )
  samples_per_slot: int = struct.field(
      pytree_node=False,
  )

  def copy_to_host_async(self: "ResultTokens") -> None:
    """Copy to host asynchronously."""
    # Do nothing for np array
    if isinstance(self.data, np.ndarray):
      return
    self.data.copy_to_host_async()

  def convert_to_numpy(self: "ResultTokens") -> "ResultTokens":
    """Converts to numpy."""
    return ResultTokens(
        np.array(self.data),
        self.tokens_idx,
        self.valid_idx,
        self.length_idx,
        self.samples_per_slot,
    )

  def get_result_at_slot(self, slot: int) -> SlotData:
    """Returns the token at a given slot.

    Args:
      slot: An integer from [0, n) representing an index into the batch.

    Note: implementations of this method must correctly handle
    microbatches, if microbatches are used.
    """
    # Potentially get multiple beams for given slot.
    start_idx = slot * self.samples_per_slot
    end_idx = (slot + 1) * self.samples_per_slot
    # Mask out any non valid tokens.
    return SlotData(
        tokens=self.data[
            start_idx:end_idx, self.tokens_idx[0] : self.tokens_idx[1]
        ],
        valid=self.data[
            start_idx:end_idx, self.valid_idx[0] : self.valid_idx[1]
        ],
        # Only get a 1D representation here
        lengths=self.data[
            start_idx:end_idx, self.length_idx[0] : self.length_idx[1]
        ][:, 0],
    )

  def get_result_at_slots(self, slots: tuple[int]) -> SlotData:
    """Returns the tokens at given slots.

    Args:
      slots: a tuple of integers from [0, n) representing indices
      into the batch.

    """
    return SlotData(
        tokens=self.data[slots, self.tokens_idx[0] : self.tokens_idx[1]],
        valid=self.data[slots, self.valid_idx[0] : self.valid_idx[1]],
        # Only get a 1D representation here
        lengths=self.data[slots, self.length_idx[0] : self.length_idx[1]][:, 0],
    )


class Engine(abc.ABC):
  """The computational core of the generative model server.

  Engine defines an API that models must adhere to as they plug into the
  JetStream efficient serving infrastructure.
  """

  @abc.abstractmethod
  def prefill(
      self,
      *,
      params: Params,
      existing_prefix: Optional[Prefix] = None,
      padded_tokens: jax.Array,
      true_length: int,
      sampler: Optional[Callable[[Any], Any]] = None,
      request_id: Optional[uuid.UUID] = None,
  ) -> Tuple[Prefix, ResultTokens]:
    """Computes a kv-cache for a set of tokens conditional on existing cache.

    existing_prefix (if provided) represents a prefix that has already been
    processed by the underlying model. tokens is logically appended
    to the text represented by `existing_prefix`. This method returns a new
    kv_cache (typically) for the resulting text.

    If sampler is passed, then the engine should use it do sample next token.
    """

  @abc.abstractmethod
  def prefill_multisampling(
      self,
      *,
      params: Params,
      existing_prefix: Optional[jax.Array] = None,
      padded_tokens: jax.Array,
      true_length: int,
      sampler: Optional[Callable[[Any], Any]] = None,  # pylint: disable=unused-argument
      rng: Optional[PRNGKeyType] = None,
      num_samples: int = 1,
  ) -> Tuple[Prefix, ResultTokens]:
    """Computes a kv-cache for a new generate request.

    With multi-sampling, the engine will generate multiple first tokens in the
    prefilling stage. The number of tokens is specified by num_samples.
    """

  @abc.abstractmethod
  def generate(
      self,
      params: Params,
      decode_state: DecodeState,
      sampler: Optional[Callable[[Any], Any]] = None,
  ) -> Tuple[DecodeState, ResultTokens]:
    """Generates tokens for each sequence being decoded in parallel.

    Generate takes a batch of pre-computed kv-caches, and computes:
      - the predicted next token for each of the sequences
      - an updated set of kv-caches

    In the case of pipelining, this will handle N cycles (where each cycle
    consists of each microbatch progressing through every stage), in
    non-pipelined code this is a full forward pass. In both cases, this accounts
    for a full embed-layerstack-unembed-sample operation.

    If sampler is passed, then the engine should use it do sample next token.
    """

  @abc.abstractmethod
  def insert(
      self,
      prefix: Prefix,
      decode_state: DecodeState,
      slot: int,
      request_id: Optional[uuid.UUID] = None,
  ) -> DecodeState:
    """Adds `new_request` into `caches` at 'slot'.

    When decoding multiple requests in parallel, when one request finishes, a
    new request must be slotted into the recently vacated spot: `insert`!

    This can occur in between and async to generate calls, and takes a lock over
    that row of the cache.

    The slot may represent a tuple of positions (e.g. microbatch, pipeline stage
    and batch), but at the engine interface level all of these are exposed as
    a [0, n) range of slots and converted internally.
    """

  @abc.abstractmethod
  def bulk_insert(
      self,
      prefix: Prefix,
      decode_state: DecodeState,
      slots: list[int],
  ) -> DecodeState:
    """Insert a single computed prefill cache into multiple slots in
    KV cache.
    """

  def free_resource(
      self,
      slot: int,  # pylint: disable=unused-argument
  ) -> Any:
    """Free cache and other decode resource for the slot.

    This function is needed for advanced attetnion kenel like PageAttetion.
    After finishing one request, the engine need to free all used page block
    resource and reuse for coming requests.
    """
    return None

  @abc.abstractmethod
  def load_params(self, *args, **kwargs) -> Params:
    """Loads parameters.

    May not be used in full production form, where weights are part of the saved
    model.
    """

  @abc.abstractmethod
  def get_prefix_destination_sharding(self) -> Any:
    """Returns the shardings necessary to transfer data between engines."""

  @abc.abstractmethod
  def get_tokenizer(
      self,
  ) -> tokenizer_pb2.TokenizerParameters:
    """Returns the info to construct a tokenizer in py/c++."""

  def build_tokenizer(
      self,
      metadata: tokenizer_pb2.TokenizerParameters,
  ) -> Tokenizer:
    """Builds a new tokenizer object and returns it."""
    return token_utils.SentencePieceTokenizer(metadata)

  @abc.abstractmethod
  def init_decode_state(self, *args, **kwargs) -> DecodeState:
    """Initialises any state which a generation step transforms."""

  @property
  @abc.abstractmethod
  def max_concurrent_decodes(self) -> int:
    """Total capacity."""

  @property
  @abc.abstractmethod
  def samples_per_slot(self) -> int:
    """Total samples per slot."""

  @property
  @abc.abstractmethod
  def max_prefill_length(self) -> int:
    """Maximum prefill length."""

  @property
  @abc.abstractmethod
  def mesh(self) -> jax.sharding.Mesh:
    """Mesh which the engine is running on."""

  @property
  @abc.abstractmethod
  def colocated_cpus(self) -> Union[list[CpuDevices], None]:
    """CPU devices colocated with the engine's accelerators."""


class JetStreamEngine(Engine):
  """A wrapper engine of the Engine class.

  JetStreamEngine defines the warmed up model server engine.
  """

  def __init__(self, downstream_engine: Engine):
    self._downstream_engine = downstream_engine

    self.prefill_buckets = None
    self.warm = False

  def prefill(
      self,
      *,
      params: Params,
      existing_prefix: Optional[Prefix] = None,
      padded_tokens: jax.Array,
      true_length: int,
  ) -> Tuple[Prefix, ResultTokens]:

    prefill_result, first_token = self._downstream_engine.prefill(
        params=params,
        padded_tokens=padded_tokens,
        true_length=true_length,
    )
    return prefill_result, first_token

  def prefill_multisampling(
      self,
      *,
      params: Params,
      existing_prefix: Optional[jax.Array] = None,
      padded_tokens: jax.Array,
      true_length: int,
      sampler: Optional[Callable[[Any], Any]] = None,  # pylint: disable=unused-argument
      rng: Optional[PRNGKeyType] = None,
      num_samples: int = 1,
  ) -> Tuple[Prefix, ResultTokens]:

    prefill_result, first_token = self._downstream_engine.prefill_multisampling(
        params=params,
        existing_prefix=existing_prefix,
        padded_tokens=padded_tokens,
        true_length=true_length,
        sampler=sampler,
        rng=rng,
        num_samples=num_samples,
    )
    return prefill_result, first_token

  def insert(
      self,
      prefix: Prefix,
      decode_state: DecodeState,
      slot: int,
      request_id: Optional[uuid.UUID] = None,
  ) -> DecodeState:

    decode_state = self._downstream_engine.insert(
        prefix=prefix,
        decode_state=decode_state,
        slot=slot,
        request_id=request_id,
    )
    return decode_state

  def bulk_insert(
      self,
      prefix: Prefix,
      decode_state: DecodeState,
      slots: list[int],
  ) -> DecodeState:

    decode_state = self._downstream_engine.bulk_insert(
        prefix=prefix,
        decode_state=decode_state,
        slots=slots,
    )
    return decode_state

  def generate(
      self, params: Params, decode_state: DecodeState
  ) -> Tuple[DecodeState, ResultTokens]:
    decode_state, sampled_tokens = self._downstream_engine.generate(
        params=params, decode_state=decode_state
    )
    return decode_state, sampled_tokens

  def load_params(self, *args, **kwargs) -> Params:
    return self._downstream_engine.load_params(*args, **kwargs)

  def get_prefix_destination_sharding(self) -> Any:
    return self._downstream_engine.get_prefix_destination_sharding()

  def get_tokenizer(
      self,
  ) -> tokenizer_pb2.TokenizerParameters:
    return self._downstream_engine.get_tokenizer()

  def build_tokenizer(
      self,
      metadata: tokenizer_pb2.TokenizerParameters,
  ) -> Tokenizer:
    """Builds a new tokenizer object and returns it."""
    return self._downstream_engine.build_tokenizer(metadata)

  def init_decode_state(self, *args, **kwargs) -> DecodeState:
    return self._downstream_engine.init_decode_state(*args, **kwargs)

  @property
  def max_concurrent_decodes(self) -> int:
    return self._downstream_engine.max_concurrent_decodes

  @property
  def samples_per_slot(self) -> int:
    return self._downstream_engine.samples_per_slot

  @property
  def max_prefill_length(self) -> int:
    return self._downstream_engine.max_prefill_length

  @property
  def mesh(self) -> jax.sharding.Mesh:
    return self._downstream_engine.mesh

  @property
  def colocated_cpus(self) -> Union[list[CpuDevices], None]:
    return self._downstream_engine.colocated_cpus

  @property
  def use_chunked_prefill(self) -> bool:
    return self._downstream_engine.use_chunked_prefill

  @property
  def prefill_chunk_size(self) -> int:
    """Maximum prefill length."""
    return self._downstream_engine.prefill_chunk_size



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests http server end-to-end."""

import subprocess
import sys
import time
import unittest


import requests


class HTTPServerTest(unittest.IsolatedAsyncioTestCase):

  @classmethod
  def setUpClass(cls):
    """Sets up a JetStream http server for unit tests."""
    cls.base_url = "http://localhost:8080"
    cls.server = subprocess.Popen(
        [
            "python",
            "-m",
            "jetstream.entrypoints.http.api_server",
            "--config=InterleavedCPUTestServer",
        ],
        stdout=sys.stdout,
        stderr=sys.stderr,
    )
    time.sleep(10)

  @classmethod
  def tearDownClass(cls):
    """Stop the server gracefully."""
    cls.server.terminate()

  async def test_root_endpoint(self):
    response = requests.get(self.base_url + "/", timeout=5)
    assert response.status_code == 200
    expected_data = {"message": "JetStream HTTP Server"}
    assert response.json() == expected_data

  async def test_health_endpoint(self):
    response = requests.get(self.base_url + "/v1/health", timeout=5)
    assert response.status_code == 200
    data = response.json()
    assert "is_live" in data
    assert data["is_live"] == "True"

  async def test_generate_endpoint(self):
    # Prepare a sample request (replace with actual data)
    sample_request_data = {
        "max_tokens": 10,
        "text_content": {"text": "translate this to french: hello world"},
    }

    response = requests.post(
        self.base_url + "/v1/generate",
        json=sample_request_data,
        stream=True,
        timeout=5,
    )
    assert response.status_code == 200
    full_response = []
    for chunk in response.iter_content(
        chunk_size=None
    ):  # chunk_size=None for complete lines
      if chunk:
        stream_response = chunk.decode("utf-8")
        print(f"{stream_response=}")
        full_response.append(stream_response)
    assert len(full_response) == 11  # 10 tokens + eos token



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .llama import LlamaForCausalLM, LlamaModel
from .management.registry import ModelSource, ModelRegistry
from .sampling.sampler import SamplingParams
from .postprocess import ModelOutput



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import dataclasses


class ModelId:
  llama_2_7b_chat_hf = "meta-llama/Llama-2-7b-chat-hf"
  llama_2_70b_chat_hf = "meta-llama/Llama-2-70b-chat-hf"


@dataclasses.dataclass
class InferenceParams:
  model_id: str
  batch_size: int
  max_seq_length: int
  max_input_length: int
  prefill_chunk_sizes: list[int]
  page_size: int
  hbm_utilization: float


class Config:
  _configs = {
      ModelId.llama_2_7b_chat_hf: InferenceParams(
          model_id=ModelId.llama_2_7b_chat_hf,
          batch_size=320,
          max_seq_length=2048,
          max_input_length=1024,
          prefill_chunk_sizes=[128, 256, 512, 1024],
          page_size=128,
          hbm_utilization=0.875,
      ),
      ModelId.llama_2_70b_chat_hf: InferenceParams(
          model_id=ModelId.llama_2_70b_chat_hf,
          batch_size=100,
          max_seq_length=2048,
          max_input_length=1024,
          prefill_chunk_sizes=[128, 256, 512, 1024],
          page_size=128,
          hbm_utilization=0.875,
      ),
  }

  @classmethod
  def get(cls, model_id: str) -> InferenceParams:
    return cls._configs[model_id]



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Inference microbenchmark for prefill and autoregressive steps."""
import datetime
import jax
import json

from absl import app
from collections.abc import MutableMapping

from jetstream.engine import token_utils

import max_utils
import maxengine
import maxtext_utils
import prefix_cache
import profiler
import pyconfig

import warnings

warnings.simplefilter("ignore", category=FutureWarning)

_WARMUP_ITERS = 2
_FLATTEN_MICROBENCHMARK_RESULTS = False
# pylint: disable=too-many-positional-arguments


def prefix_cache_benchmark(
    prefix, prefill_length: int, true_length: int, common_prefix_proportion: float, prefix_cache_entries_num: int, iters: int
):
  """Handles running prefix cache benchmark, and printing results.

  Create different key with half of prefill_length common prefix insert into cache.
  The value is not relevant to the cache for now. Just copy the prefix for every cache entry.
  1. Fill the prefix cache to full capacity.
  2. Benchmark save prefix cache with evicting time average by prefix_cache_entries_num.
  3. Benchmark fetch_longest_common_prefix_key average by iters.
  4. Benchmark load prefix cache time average by iters.

  Args:
    prefix: prefix return from prefill function
    prefill_length: prefill token length after padding
    true_length: true prefill token length
    common_prefix_proportion: [0., 1.] common prefix proportion to the prefill_length
    prefix_cache_entries_num: number of prefix cache entries insert into PrefixCache
    iters: repeat time to test fetch_longest_common_prefix_key and load from cache
  """

  print(f"Prefix Cache benchmark results for prefill length {prefill_length}:\n")

  value: prefix_cache.Value = prefix_cache.Value(
      prefix=prefix,
      true_length=true_length,
      padded_length=prefill_length,
      tokens=tuple(i for i in range(prefill_length)),
  )

  def copy_jax_array(x):
    return x.copy()

  def clone_value():
    return prefix_cache.Value(
        prefix=jax.tree.map(copy_jax_array, value.prefix),
        true_length=value.true_length,
        padded_length=value.padded_length,
        tokens=value.tokens,
        prefix_size_bytes=value.prefix_size_bytes,
        device=value.device,
    )

  prefix_size_bytes_gb = value.prefix_size_bytes / 1024 / 1024 / 1024
  max_bytes = prefix_cache_entries_num * value.prefix_size_bytes
  # TODO(yuyanpeng): test hierarchical cache
  prefix_cache_inst = prefix_cache.PrefixCache(hbm_bytes=max_bytes, dram_bytes=max_bytes)
  common_len = int(prefill_length * common_prefix_proportion)
  remain_len = prefill_length - common_len
  common_prefix_key = tuple(i for i in range(common_len))

  # Fill the prefix caching
  new_value_list = []
  for c_idx in range(prefix_cache_entries_num):
    # Add 100 to make sure filled prefix caching will not share the common_prefix_key.
    # The later save prefix part will evict all of them.
    key = tuple(100 + i + c_idx * prefill_length for i in range(prefill_length))
    new_value = clone_value()
    prefix_cache_inst.save(key, new_value)
    new_value_list.append(new_value)
  jax.block_until_ready(new_value_list)
  del new_value_list

  # Save prefix
  new_value = None
  save_sec = 0
  for c_idx in range(iters):
    key = common_prefix_key + tuple(i + c_idx * remain_len for i in range(remain_len))
    # values are not relevant for caching now, just clone the same tokens and values for test
    new_value = clone_value()
    jax.block_until_ready(new_value)
    start = datetime.datetime.now()
    prefix_cache_inst.save(key, new_value)
    end = datetime.datetime.now()
    save_sec += (end - start).total_seconds()
  del new_value
  save_avg_ms = save_sec * 1000 / iters

  # Fetch longest prefix key
  key_load = common_prefix_key + tuple(i + prefix_cache_entries_num * remain_len for i in range(remain_len))
  matched_key = None
  fetch_sec = 0
  for _ in range(iters):
    start = datetime.datetime.now()
    matched_key = prefix_cache_inst.fetch_longest_common_prefix_key(key_load)
    end = datetime.datetime.now()
    fetch_sec += (end - start).total_seconds()
  fetch_avg_ms = fetch_sec * 1000 / iters

  assert matched_key is not None

  # Load prefix
  load_sec = 0
  value_load = None
  for _ in range(iters):
    start = datetime.datetime.now()
    loaded_value = prefix_cache_inst.load(matched_key)
    jax.block_until_ready(loaded_value)
    end = datetime.datetime.now()
    load_sec += (end - start).total_seconds()
  del value_load
  load_avg_ms = load_sec * 1000 / iters

  print(
      f"PrefixCaching results:\n"
      f"\tPer prefix size bytes: {prefix_size_bytes_gb:.3f} GB\n"
      f"\tAverage save cache time: {save_avg_ms:.3f} ms\n"
      f"\tAverage fetch longest prefix time: {fetch_avg_ms:.3f} ms\n"
      f"\tAverage load cache time: {load_avg_ms:.3f} ms\n\n\n"
  )
  del prefix_cache_inst


def prefill_benchmark_loop(engine_prefill, params, tokens, true_length, iters, num_samples: int | None = None):
  """Inner loop for benchmarking prefill step."""
  start = datetime.datetime.now()
  rng = jax.random.PRNGKey(1234)
  prefill_result = None
  for _ in range(iters):
    rng, rng_prefill = jax.random.split(rng)
    if num_samples is None:
      prefill_result, _ = engine_prefill(params, tokens, true_length, rng_prefill)
    else:
      prefill_result, _ = engine_prefill[num_samples](params, tokens, true_length, rng_prefill, None)
  jax.block_until_ready(prefill_result)
  end = datetime.datetime.now()
  del prefill_result
  return (end - start).total_seconds()


def prefill_benchmark(config, engine_prefill, params, tokens, true_length, num_model_params, iters):
  """Handles warmup, running prefill benchmark, and printing results."""
  rng = jax.random.PRNGKey(1234)
  prefill_result = None
  for _ in range(_WARMUP_ITERS):
    rng, rng_prefill = jax.random.split(rng)
    prefill_result, _ = engine_prefill(params, tokens, true_length, rng_prefill)
  jax.block_until_ready(prefill_result)
  del prefill_result

  print(f"Prefill benchmark results for length {tokens.size}:\n")
  time_in_s = prefill_benchmark_loop(engine_prefill, params, tokens, true_length, iters)
  prefill_average_ms = 1000 * time_in_s / iters
  prefill_tflops_per_device, _, _ = maxtext_utils.calculate_prefill_tflops_per_device(num_model_params, tokens.size, config)
  tflops_per_sec_per_device = prefill_tflops_per_device / prefill_average_ms * 1000.0
  print(
      f"\tPrefill step average time: {prefill_average_ms:.3f} ms\n"
      f"\tPrefill total TFLOPs/device: {prefill_tflops_per_device:.3f}\n"
      f"\tPrefill TFLOPs/sec/device: {tflops_per_sec_per_device:.3f}\n\n\n\n"
  )
  result_dict = {
      "time_in_ms": prefill_average_ms,
      "total_tflops_per_device": prefill_tflops_per_device,
      "tflops_per_sec_per_device": tflops_per_sec_per_device,
  }
  return result_dict


def prefill_multisampling_benchmark(config, engine_prefill_multisampling, params, tokens, true_length, iters):
  """Handles warmup, running prefill benchmark, and printing results."""
  rng = jax.random.PRNGKey(1234)
  prefill_result = None
  for _ in range(_WARMUP_ITERS):
    rng, rng_prefill = jax.random.split(rng)
    for num_samples in config.inference_microbenchmark_num_samples:
      prefill_result, _ = engine_prefill_multisampling[num_samples](params, tokens, true_length, rng_prefill, None)
  jax.block_until_ready(prefill_result)
  del prefill_result

  print(f"Multi-sampling prefill benchmark results for length {tokens.size}:\n")
  result_dict = {}
  for num_samples in config.inference_microbenchmark_num_samples:
    time_in_s = prefill_benchmark_loop(engine_prefill_multisampling, params, tokens, true_length, iters, num_samples)
    multisampling_prefill_average_ms = 1000 * time_in_s / iters
    print(
        f"\nNum samples: {num_samples}\n" f"\tPrefill step average time: {multisampling_prefill_average_ms:.3f} ms\n\n\n\n"
    )
    result_dict[num_samples] = {
        "time_in_ms": multisampling_prefill_average_ms,
    }
  return result_dict


def prefill_insert_benchmark_loop(
    config, engine_insert, decode_state, params, total_slots, tokens, true_length, iters, profile_name
):
  """Inner loop for benchmarking prefill and insert step."""
  prof = profiler.Profiler(config)
  prof.activate(optional_postfix=profile_name)
  start = datetime.datetime.now()
  rng = jax.random.PRNGKey(1234)
  for i in range(iters):
    rng, rng_prefill = jax.random.split(rng)
    decode_state = engine_insert(tokens, true_length, rng_prefill, decode_state, int(i % total_slots), params)
  jax.block_until_ready(decode_state)
  end = datetime.datetime.now()
  prof.deactivate()
  return (end - start).total_seconds(), decode_state


def prefill_insert_benchmark(config, engine_insert, decode_state, params, total_slots, tokens, true_length, iters):
  """Handles warmup, running insert benchmark, and printing results."""
  rng = jax.random.PRNGKey(1234)
  for i in range(_WARMUP_ITERS):
    rng, rng_prefill = jax.random.split(rng)
    decode_state = engine_insert(tokens, true_length, rng_prefill, decode_state, int(i % total_slots), params)
  jax.block_until_ready(decode_state)

  print(f"Prefill and insert benchmark results for length {tokens.size}:\n")
  time_in_s, decode_state = prefill_insert_benchmark_loop(
      config, engine_insert, decode_state, params, total_slots, tokens, true_length, iters, f"prefill_insert_{tokens.size}"
  )
  prefill_insert_average_ms = time_in_s / iters * 1000.0
  print(f"\tPrefill + Insert step average time: {prefill_insert_average_ms:.3f} ms\n\n\n\n")
  result_dict = {"time_in_ms": prefill_insert_average_ms}
  return result_dict, decode_state


def ar_benchmark_loop(config, engine_generate, params, decode_state, iters, profile_name):
  """Inner loop for benchmarking ar step."""
  prof = profiler.Profiler(config)
  prof.activate(optional_postfix=profile_name)
  start = datetime.datetime.now()
  rng = jax.random.PRNGKey(1234)
  for _ in range(iters):
    rng, rng_generate = jax.random.split(rng)
    decode_state, _ = engine_generate(params, decode_state, rng_generate)
  jax.block_until_ready(decode_state)
  end = datetime.datetime.now()
  prof.deactivate()
  return (end - start).total_seconds(), decode_state


def ar_benchmark(config, engine_generate, params, decode_state, global_batch_size, cache_size, model_size, iters):
  """Handles warmup, running ar benchmark, and printing results."""
  rng = jax.random.PRNGKey(1234)
  for _ in range(_WARMUP_ITERS):
    rng, rng_generate = jax.random.split(rng)
    decode_state, _ = engine_generate(params, decode_state, rng_generate)
  jax.block_until_ready(decode_state)

  time_in_s, decode_state = ar_benchmark_loop(
      config, engine_generate, params, decode_state, iters, profile_name="autoregress"
  )
  seconds_per_step = time_in_s / iters
  ar_average_ms = seconds_per_step * 1000
  total_throughput = global_batch_size / seconds_per_step

  GB_per_step_per_device = (model_size + cache_size) / 1e9 / jax.device_count()
  bw_per_device = GB_per_step_per_device / seconds_per_step
  print(
      f"AutoRegressive results:\n"
      f"\tAR step average time: {ar_average_ms:.3f} ms\n"
      f"\tAR step average time per seq: {ar_average_ms/global_batch_size:.3f} ms\n"
      f"\tAR global batch size: {global_batch_size}\n"
      f"\tAR throughput: {total_throughput:.3f} tokens/second\n"
      f"\tAR memory bandwidth per device: {bw_per_device:.3f} GB/s\n\n\n"
  )

  result_dict = {
      "step_in_ms": ar_average_ms,
      "step_in_ms_per_seq": ar_average_ms / global_batch_size,
      "global_batch_size": global_batch_size,
      "total_throughput_tokens_per_second": total_throughput,
      "bw_per_device_GB_per_second": bw_per_device,
  }
  return result_dict, decode_state


def collate_results(config, results, model_size, cache_size, num_model_params, incl_config=False):
  """Adds model/cache size info and optionally config info to results."""
  results["sizes"] = {
      "model_size_in_gb": model_size / 1e9,
      "cache_size_in_gb": cache_size / 1e9,
      "model_params_in_billions": num_model_params / 1e9,
  }
  if incl_config:
    results["config"] = {}
    for k, v in dict(config.get_keys()).items():
      results["config"][k] = str(v) if k == "dtype" else v  # json fails with original dtype
  return results


def flatten_dict(dictionary, prefix="", sep="_"):
  results = []
  for k, v in dictionary.items():
    new_key = str(prefix) + sep + str(k) if prefix else k
    if isinstance(v, MutableMapping):
      results.extend(flatten_dict(v, new_key, sep=sep).items())
    else:
      results.append((new_key, v))
  return dict(results)


def write_results(results, filename, flatten_microbenchmark_results):
  """Write the results microbenchmark results to a json file."""
  if flatten_microbenchmark_results:
    results["flattened_results"] = flatten_dict(results)
  if filename:
    with open(filename, "w", encoding="utf-8") as f:
      json.dump(results, f, indent=2)
  return results


def print_results_for_analyze(results):
  """Print results."""
  print("\nFor usage in analyze_sharegpt.py :")

  if "prefill" in results:
    prefill_bucket_size_to_ms = {}
    for k, v in results["prefill"].items():
      prefill_bucket_size_to_ms[int(k)] = round(v["time_in_ms"], 3)
    print(f"PREFILL_BUCKET_SIZE_TO_MS = {prefill_bucket_size_to_ms}")

  if "prefill-multisampling" in results:
    multi_sampling_prefill_bucket_size_to_ms = {}
    for prefill_length, result_dict in results["prefill-multisampling"].items():
      multi_sampling_prefill_bucket_size_to_ms[int(prefill_length)] = {}
      for num_samples, v in result_dict.items():
        multi_sampling_prefill_bucket_size_to_ms[int(prefill_length)][num_samples] = round(v["time_in_ms"], 3)
    print(f"MULTISAMPLING_PREFILL_BUCKET_SIZE_TO_MS = {multi_sampling_prefill_bucket_size_to_ms}")

  if "insert" in results:
    insert_bucket_size_to_ms = {}
    for k, v in results["insert"].items():
      insert_bucket_size_to_ms[int(k)] = round(v["time_in_ms"], 3)
    print(f"INSERT_BUCKET_SIZE_TO_MS = {insert_bucket_size_to_ms}")

  if "autoregressive" in results:
    print(f"SYSTEM_TIME_PER_DECODE_TOKEN_MS = {results['autoregressive']['step_in_ms_per_seq']}")


def summarize_prefill_result(engine_prefill, params, tokens, true_length):
  """Summarize Prefill result."""
  print(f"Prefill result of length {tokens.size}:\n")
  rng = jax.random.PRNGKey(1234)
  prefill_result, _ = engine_prefill(params, tokens, true_length, rng)
  jax.block_until_ready(prefill_result)
  num_prefill_logits_params, total_prefill_logits_size, avg_prefill_logits_param_size = max_utils.summarize_pytree_data(
      prefill_result["logits"], name="Prefill Logits", raw=True
  )
  num_prefill_cache_params, total_prefill_cache_size, avg_prefill_cache_param_size = max_utils.summarize_pytree_data(
      prefill_result["cache"], name="Prefill Cache"
  )
  del prefill_result
  return {
      "num_logits_params": num_prefill_logits_params,
      "total_logits_size": total_prefill_logits_size,
      "avg_logits_param_size": avg_prefill_logits_param_size,
      "num_cache_params": num_prefill_cache_params,
      "total_cache_size": total_prefill_cache_size,
      "avg_cache_param_size": avg_prefill_cache_param_size,
  }


def run_benchmarks(config):
  """Run microbenchmarks."""
  engine = maxengine.MaxEngine(config)
  rng = jax.random.PRNGKey(1234)
  rng, rng_load_params = jax.random.split(rng)
  params = engine.load_params(rng_load_params)
  prefill_lengths = [int(l) for l in config.inference_microbenchmark_prefill_lengths.split(",")]
  stages_to_benchmark = config.inference_microbenchmark_stages.split(",")
  benchmark_loop_iters = config.inference_microbenchmark_loop_iters

  text = config.prompt
  metadata = engine.get_tokenizer()
  vocab = token_utils.load_vocab(metadata.path, metadata.extra_ids)
  rng, rng_init_decode = jax.random.split(rng)

  generate_executable, params, decode_state_executable = engine.aot_compile(params, pass_rng_shape=True)
  decode_state = decode_state_executable(rng_init_decode)

  _, cache_size, _ = max_utils.summarize_pytree_data(decode_state["cache"], name="Cache")
  num_model_params, model_size, _ = max_utils.summarize_pytree_data(params, name="Model")

  benchmark_results = {}
  if "prefill" in stages_to_benchmark:
    benchmark_results["prefill-result-sizes"] = {}
    benchmark_results["prefill"] = {}
    benchmark_results["insert"] = {}
    prefill_tokens = {}
    prefill_true_lengths = {}
    prefill_executable = {}
    prefill_insert_executable = {}
    i32_scalar = jax.ShapeDtypeStruct((), int)
    rng_shape = jax.ShapeDtypeStruct([4], jax.numpy.dtype("uint32"))

    for prefill_length in prefill_lengths:
      prefill_tokens[prefill_length], prefill_true_lengths[prefill_length] = token_utils.tokenize_and_pad(
          text, vocab, is_bos=True, prefill_lengths=[prefill_length]
      )

      key_shape = jax.ShapeDtypeStruct([prefill_length], jax.numpy.dtype("int32"))
      prefill_executable[prefill_length] = (
          jax.jit(
              engine.prefill_aot,
              in_shardings=(engine.param_layouts, None, None, None),
          ).lower(params, key_shape, i32_scalar, rng_shape)
      ).compile(compiler_options=None)

      prefill_insert_executable[prefill_length] = (
          jax.jit(
              engine.prefill_insert,
              in_shardings=(None, None, None, engine.decode_state_layouts, None, engine.param_layouts),
              out_shardings=(engine.decode_state_layouts),
              donate_argnames=("decode_state",),
          ).lower(key_shape, i32_scalar, rng_shape, engine.decode_state_shapes, i32_scalar, params)
      ).compile(compiler_options=None)

      benchmark_results["prefill-result-sizes"][prefill_length] = summarize_prefill_result(
          prefill_executable[prefill_length], params, prefill_tokens[prefill_length], prefill_true_lengths[prefill_length]
      )

    if "prefix_cache" in stages_to_benchmark:
      for prefill_length in prefill_lengths:
        rng_cache = jax.random.PRNGKey(1234)
        prefill_result, _ = prefill_executable[prefill_length](
            params, prefill_tokens[prefill_length], prefill_true_lengths[prefill_length], rng_cache
        )
        prefix_cache_benchmark(
            prefill_result,
            prefill_length,
            prefill_true_lengths[prefill_length],
            config.inference_microbenchmark_prefix_cache_common_prefix_proportion,
            config.inference_microbenchmark_prefix_cache_entries_num,
            benchmark_loop_iters,
        )
        del prefill_result

    for prefill_length in prefill_lengths:
      benchmark_results["prefill"][prefill_length] = prefill_benchmark(
          config,
          prefill_executable[prefill_length],
          params,
          prefill_tokens[prefill_length],
          prefill_true_lengths[prefill_length],
          num_model_params,
          benchmark_loop_iters,
      )

      prefill_insert_time, decode_state = prefill_insert_benchmark(
          config,
          prefill_insert_executable[prefill_length],
          decode_state,
          params,
          engine.max_concurrent_decodes,
          prefill_tokens[prefill_length],
          prefill_true_lengths[prefill_length],
          benchmark_loop_iters,
      )
      benchmark_results["insert"][prefill_length] = {}
      benchmark_results["insert"][prefill_length]["time_in_ms"] = (
          prefill_insert_time["time_in_ms"] - benchmark_results["prefill"][prefill_length]["time_in_ms"]
      )

  if "prefill-multisampling" in stages_to_benchmark:
    benchmark_results["prefill-multisampling"] = {}
    multisampling_prefill_executable = {}
    i32_scalar = jax.ShapeDtypeStruct((), int)
    rng_shape = jax.ShapeDtypeStruct([4], jax.numpy.dtype("uint32"))
    # Compile the program in advance.
    for prefill_length in prefill_lengths:
      key_shape = jax.ShapeDtypeStruct([prefill_length], jax.numpy.dtype("int32"))
      multisampling_prefill_executable[prefill_length] = {}
      for num_samples in config.inference_microbenchmark_num_samples:
        multisampling_prefill_executable[prefill_length][num_samples] = (
            jax.jit(
                engine.prefill_multisampling_aot,
                in_shardings=(engine.param_layouts, None, None, None, None),
                static_argnames=("num_samples",),
            ).lower(params, key_shape, i32_scalar, rng_shape, num_samples, None)
        ).compile(compiler_options=None)

    for prefill_length in prefill_lengths:
      benchmark_results["prefill-multisampling"][prefill_length] = prefill_multisampling_benchmark(
          config,
          multisampling_prefill_executable[prefill_length],
          params,
          prefill_tokens[prefill_length],
          prefill_true_lengths[prefill_length],
          benchmark_loop_iters,
      )

  if "generate" in stages_to_benchmark:
    benchmark_results["autoregressive"], decode_state = ar_benchmark(
        config,
        generate_executable,
        params,
        decode_state,
        engine.max_concurrent_decodes,
        cache_size,
        model_size,
        benchmark_loop_iters,
    )

  results = collate_results(config, benchmark_results, model_size, cache_size, num_model_params)
  print_results_for_analyze(results)
  if config.inference_microbenchmark_log_file_path:
    write_results(
        results,
        filename=config.inference_microbenchmark_log_file_path,
        flatten_microbenchmark_results=_FLATTEN_MICROBENCHMARK_RESULTS,
    )
  return results


def main(argv):
  jax.config.update("jax_default_prng_impl", "unsafe_rbg")
  run_benchmarks(pyconfig.initialize(argv))


if __name__ == "__main__":
  app.run(main)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os

import pandas as pd

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("dataset.py")


class Dataset:

  def __init__(
      self,
      dataset_path: str,
      input_mode: str,
      total_sample_count: int = 24576,
      perf_count_override: int = 0,
      dataset_rename_cols: str = "",
  ):
    if not os.path.isfile(dataset_path):
      log.warn(
          "Processed pickle file {} not found. Please check that the path is correct".format(
              dataset_path
          )
      )
    self.dataset_path = dataset_path

    self._input_mode = validate_sample_mode(input_mode)
    self.dataset_rename_cols = dataset_rename_cols
    self.load_processed_dataset()

    self.total_sample_count = min(len(self.input_ids_strs), total_sample_count)
    self.perf_count = perf_count_override or self.total_sample_count

  @property
  def input_ids_strs(self):
    return self._input_ids_strs

  @property
  def input_texts(self):
    return self._input_texts

  @property
  def input_token_lengths(self):
    return self._input_token_lengths

  @property
  def inputs(self):
    return self._inputs

  @property
  def inputs_with_token_lengths(self):
    return self._inputs_with_token_lengths

  def load_processed_dataset(self):
    processed_data = pd.read_pickle(self.dataset_path)
    if self.dataset_rename_cols:
      rename_dict = json.loads(self.dataset_rename_cols)
      processed_data.rename(columns=rename_dict, inplace=True)
      log.info(f"Renaming columns of dataset with mapping: {rename_dict}")

    self._input_ids_strs = []
    for input_ids in processed_data["tok_input"]:
      input_ids_str = ",".join([str(input_id) for input_id in input_ids])
      self._input_ids_strs.append(input_ids_str)

    self._input_texts = []
    for input_text in processed_data["input"]:
      self._input_texts.append(input_text)

    self._input_token_lengths = []
    for token_length in processed_data["tok_input_length"]:
      self._input_token_lengths.append(token_length)

    log.info(f"input_mode is {self._input_mode}")
    self._inputs = (
        self._input_ids_strs
        if self._input_mode == "tokenized"
        else self._input_texts
    )
    log.info(f"example sample input is {self._inputs[0]}")
    self._inputs_with_token_lengths = [
        (input_ids_str_or_input_text, token_length)
        for input_ids_str_or_input_text, token_length in zip(
            self._inputs, self._input_token_lengths
        )
    ]

  def LoadSamplesToRam(self, sample_list):
    pass

  def UnloadSamplesFromRam(self, sample_list):
    pass

  def __del__(self):
    pass


SAMPLE_MODE_CHOICES = ["tokenized", "text"]


def validate_sample_mode(sample_mode: str) -> str:
  if sample_mode not in SAMPLE_MODE_CHOICES:
    raise ValueError(
        "The sample_mode should be set to either `tokenized` or `text`."
    )
  return sample_mode



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Request and response classes"""

# TODO: Clean up and merge the request to improve the readability.

import asyncio
from dataclasses import dataclass, field
import numpy as np
import jax
import threading
import queue


@dataclass
class Response:
  generated_text: str
  generated_tokens: list[int] | int
  input_tokens: list[int] | None = None


@dataclass
class OnlineRequest:
  prompt: str
  res_queue: asyncio.Queue[Response]


@dataclass
class OfflineRequest:
  prompt: str


@dataclass
class Request:
  """Request for holding the input and output information"""

  id: str
  prompt: str
  prompt_token_ids: list[int] = field(default_factory=lambda: [])
  generated_text: str = ""
  generated_token_ids: list[int] = field(default_factory=lambda: [])
  aio_response_queue: asyncio.Queue[Response] | None = None
  completed: bool = False


@dataclass
class PrefillRequest:
  """class for new request need to be processed in the prefill phase"""

  id: str
  unpadded_token_ids: list[int]
  chunk_idx: int
  chunk_size: int
  page_indices: list[int]
  device_token_ids: jax.Array
  device_positions: jax.Array


@dataclass
class GenerateRequest:
  """class for new request need to be processed in the generate phase"""

  id: str
  slot: int
  pos: int
  page_indices: list[int]
  device_prefill_token_id: jax.Array


@dataclass
class GenerateState:
  """generate phase state"""

  token_ids: jax.Array  # batch_size
  positions: jax.Array  # batch_size
  page_table: jax.Array  # batch_size, num_pages_per_seq
  available_slots: queue.SimpleQueue
  active_slot_req_map: dict[int, GenerateRequest]
  map_mutex: threading.Lock = threading.Lock()


@dataclass
class PostProcessRequest:
  """Post process request"""

  prefill_request_id: str | None
  prefill_token_id: jax.Array | np.ndarray
  prefill_done: jax.Array | np.ndarray

  generate_active_slots: list[int]
  generate_active_request_ids: list[str]
  generate_token_ids: jax.Array | np.ndarray
  generate_done: jax.Array | np.ndarray



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark JetStream online serving.

On the server side, run one of the following commands:
    * For real server, you need to pass correct server config (include the
      model config that being passed into your engine impl) to the command
      below. Refer to config_lib.py and implementations/mock/config.py for
      config impl detail.

    (run with real server)
    python -m jetstream.core.implementations.<your_impl>.server \
        --config <your_server_config>

    (run with mock server)
    python -m jetstream.core.implementations.mock.server

On the client side, run:
    * For real server and shareGPT dataset, you need to pass the tokenizer,
      server config, and dataset flags to the command below, and make some
      changes to the tokenizer logic in the benchmark script (get_tokenizer
      and sample_requests func) to use your tokenizer correctly.
    * Add `--save-result` flag to save the benchmark result to a json file in
      current folder.
    * You can also add `--run_eval true` if you want to calculate ROUGE score
      on the predicted outputs.

    (run with real model and engines)
    python -m benchmarks.benchmark_serving \
        --tokenizer <your_tokenizer> \
        --dataset <target_dataset_name> \
        --dataset-path <target_dataset_path> \
        --request-rate <request_rate>

    (run with mock)
    python -m benchmarks.benchmark_serving \
        --request-rate 1

e2e example:
python3 benchmark_serving.py \
    --tokenizer /home/{username}/maxtext/assets/tokenizer \
    --num-prompts 100 \
    --dataset sharegpt \
    --dataset-path ~/ShareGPT_V3_unfiltered_cleaned_split.json

"""


import argparse
import asyncio
from dataclasses import dataclass, field
from datetime import datetime
import gc
import json
import os
import random
import time
from typing import Any, AsyncGenerator, Optional

from benchmarks.eval_accuracy import eval_accuracy
from benchmarks.eval_accuracy_mmlu import eval_accuracy_mmlu
from benchmarks.metrics import CounterMetric, EventMetric
import grpc
from jetstream.core.proto import jetstream_pb2
from jetstream.core.proto import jetstream_pb2_grpc
from jetstream.engine.token_utils import load_vocab
from jetstream.external_tokenizers.llama3 import llama3_tokenizer
import numpy as np
import pandas
from tqdm.asyncio import tqdm  # pytype: disable=pyi-error
from transformers import AutoTokenizer


def str2bool(v: str) -> bool:
  """Convert a string of truth to True or False.

  Args:
    - v (str):
      - True values are 'y', 'yes', 't', 'true', and '1';
      - False values are 'n', 'no', 'f', 'false', and '0'.

  Returns:
    bool: True or False

  Raises:
    ValueError if v is anything else.
  """
  v = v.lower()
  true_values = ["y", "yes", "t", "true", "1"]
  false_values = ["n", "no", "f", "false", "0"]
  if v in true_values:
    return True
  elif v in false_values:
    return False
  else:
    raise ValueError(f"Invalid value '{v}'!")


class AsyncCounter:
  """An counter class for counting and quota management with asycio,
  not thread safe. It's safe with asyncio as value changes are done
  outside of await statements.
  """

  def __init__(self, init_value: int, block_on_zero_seconds=0.002):
    """
    Args:
      init_value: Initial value for the counter.
      block_on_zero_seconds: if greater than 0, the counter will spin when
        value hits 0, hence can be used for quota management.
    """
    self._init_value = init_value
    self._value = init_value
    self._block_on_zero_seconds = block_on_zero_seconds

  async def inc(self):
    self._value += 1

  async def dec(self):
    while True:
      if self._value > 0 or self._block_on_zero_seconds <= 0.0:
        self._value -= 1
        return
      await asyncio.sleep(self._block_on_zero_seconds)

  def value(self):
    return self._value

  def delta(self):
    return self._init_value - self._value


@dataclass
class BenchmarkMetrics:
  """Data class to store benchmark metrics."""

  completed: int
  total_input: int
  total_output: int
  request_throughput: float
  input_throughput: float
  output_throughput: float
  overall_throughput: float

  ttft: EventMetric  # Time-to-first-token
  ttst: EventMetric  # Time-to-second-token
  tpot: EventMetric  # Time-per-output-token


@dataclass
class InputRequest:
  prompt: str = ""
  prompt_len: int = 0
  output: str = ""
  output_len: int = 0
  sample_idx: int = -1


@dataclass
class RequestFuncOutput:
  """Data class to store the response of a request."""

  input_request: Optional[InputRequest] = None
  generated_token_list: list[int] = field(default_factory=list)
  generated_text: str = ""
  success: bool = False
  latency_sec: float = 0
  ttft_sec: float = 0
  ttst_sec: float = 0
  prompt_len: int = 0

  # Flatten the structure and return only the necessary results
  def to_dict(self):
    if self.input_request:
      prompt = self.input_request.prompt
      original_output = self.input_request.output
      sample_idx = self.input_request.sample_idx
    else:
      prompt = None
      original_output = None
      sample_idx = None
    return {
        "prompt": prompt,
        "original_output": original_output,
        "generated_text": self.generated_text,
        "success": self.success,
        "latency_sec": self.latency_sec,
        "ttft_sec": self.ttft_sec,
        "ttst_sec": self.ttst_sec,
        "prompt_len": self.prompt_len,
        "sample_idx": sample_idx,
    }


def get_tokenizer(
    model_id: str,
    tokenizer_name: str,
    use_hf_tokenizer: bool,
) -> Any:
  """Return a tokenizer or a tokenizer placholder."""
  if tokenizer_name == "test":
    print("Using test tokenizer")
    return "test"
  elif use_hf_tokenizer:
    # Please accept agreement to access private/gated models in HF, and
    # follow up instructions below to set up access token
    # https://huggingface.co/docs/transformers.js/en/guides/private
    print(f"Using HuggingFace tokenizer: {tokenizer_name}")
    return AutoTokenizer.from_pretrained(tokenizer_name)
  elif model_id == "llama-3":
    # Llama 3 uses a tiktoken tokenizer.
    print(f"Using llama-3 tokenizer: {tokenizer_name}")
    return llama3_tokenizer.Tokenizer(tokenizer_name)
  else:
    # Use JetStream tokenizer util. It's using the sentencepiece wrapper in
    # seqio library.
    print(f"Using tokenizer: {tokenizer_name}")
    vocab = load_vocab(tokenizer_name)
    return vocab.tokenizer


def load_sharegpt_dataset(
    dataset_path: str,
    conversation_starter: str,
) -> list[tuple[Any, Any]]:
  # Load the dataset.
  with open(dataset_path, "r", encoding="utf-8") as f:
    dataset = json.load(f)
  # Filter out the conversations with less than 2 turns.
  dataset = [data for data in dataset if len(data["conversations"]) >= 2]

  # Filter based on conversation starter
  if conversation_starter != "both":
    dataset = [
        data
        for data in dataset
        if data["conversations"][0]["from"] == conversation_starter
    ]
  # Only keep the first two turns of each conversation.
  dataset = [
      (data["conversations"][0]["value"], data["conversations"][1]["value"])
      for data in dataset
  ]

  return dataset


def load_openorca_dataset_pkl(
    dataset_path: str,
) -> list[tuple[Any, Any]]:
  if not dataset_path:
    dataset_path = "open_orca_gpt4_tokenized_llama.calibration_1000.pkl"
  # read pickle file
  samples = pandas.read_pickle(
      os.path.join(
          os.path.dirname(os.path.relpath(__file__)),
          dataset_path,
      )
  )

  prompts = []
  outputs = []
  for _, row in samples.iterrows():
    prompts.append(row["input"])
    outputs.append(row["output"])

  return [(prompt, output) for prompt, output in zip(prompts, outputs)]


def load_longcontext_dataset_pkl(
    dataset_path: str,
) -> list[tuple[Any, Any]]:
  assert os.path.isfile(dataset_path)

  # read pickle file
  data = pandas.read_pickle(dataset_path)

  samples = []
  for _, row in data.iterrows():
    samples.append((row["input"], row["ref_output"]))

  return samples


def load_mmlu_dataset_csv(dataset_path: str) -> tuple[Any, dict[str, str]]:
  assert dataset_path != ""
  dataset = []
  prompts_per_subject = dict()
  for cvs_file in os.listdir(dataset_path):
    if cvs_file.endswith(".csv"):
      subject = " ".join(cvs_file.split("_")[:-1])
      if subject not in prompts_per_subject:
        prompts_per_subject[subject] = ""
      filepath = os.path.join(dataset_path, cvs_file)
      data = pandas.read_csv(filepath, header=None)
      data["subject"] = subject
      dataset.append(data)

  combined_dataset = pandas.concat(dataset, ignore_index=True)
  header_dict = {
      0: "question",
      1: "A",
      2: "B",
      3: "C",
      4: "D",
      5: "answer",
  }
  combined_dataset.rename(columns=header_dict, inplace=True)
  return combined_dataset, prompts_per_subject


def gen_mmlu_qa(data: Any, mmlu_method: str = "") -> str:

  output = ""
  for _, row in data.iterrows():
    output += (
        f"Question: {row['question']}\n"
        f"Choices:\n"
        f"(A) {row['A']}\n"
        f"(B) {row['B']}\n"
        f"(C) {row['C']}\n"
        f"(D) {row['D']}\n"
    )

    output += "\nCorrect answer: "

    if mmlu_method == "HELM":
      output += f"({row['answer']})\n\n"
    elif mmlu_method == "Harness":
      content = row[row["answer"].upper()]
      output += f"({row['answer']}) {content}\n\n"

  return output


def gen_mmlu_prompt(
    dataset_path: str, num_shots: int = 1, mmlu_method: str = "HELM"
) -> list[tuple[Any, Any]]:
  combined_dataset, prompts_per_subject = load_mmlu_dataset_csv(dataset_path)
  num_rows, _ = combined_dataset.shape
  print(f"Loaded {num_rows} data from mmlu dataset")

  for subject in prompts_per_subject:
    header = (
        f"The following are multiple choice questions (with answers) "
        f"about {subject}:\n"
    )
    shots_data = combined_dataset[combined_dataset["subject"] == subject].head(
        num_shots
    )
    prompts_per_subject[subject] = header + gen_mmlu_qa(
        shots_data, mmlu_method=mmlu_method
    )

  mmlu_data = []
  for _, row in combined_dataset.iloc[num_shots:].iterrows():
    question_prompt = gen_mmlu_qa(pandas.DataFrame([row]))
    output = row["answer"]
    prompt = prompts_per_subject[row["subject"]] + question_prompt
    mmlu_data.append((prompt, output))

  return mmlu_data


def load_math500_dataset(dataset_path: str) -> list[tuple[Any, Any]]:
  if not dataset_path:
    dataset_path = "benchmarks/huggingfaceh4_math500.json"
  abs_path = os.path.abspath(dataset_path)
  with open(abs_path, "r", encoding="utf-8") as f:
    dataset = json.load(f)
  return [
      (data["problem"], (data["solution"], data["answer"])) for data in dataset
  ]


def tokenize_dataset(
    dataset: list[tuple[Any, Any, Any]],
    tokenizer: Any,
) -> list[tuple[str, Any, str, int, int, int]]:
  tokenized_dataset = []

  for prompt, output, idx in dataset:
    if isinstance(output, tuple):
      output_len = len(tokenizer.encode(output[0]))
      output_tokens = output[1]
    else:
      output_len = len(tokenizer.encode(output))
      output_tokens = output

    prompt_tokens = tokenizer.encode(prompt)

    tokenized_data = (
        prompt,
        prompt_tokens,
        output_tokens,
        len(prompt_tokens),
        output_len,
        idx,
    )
    tokenized_dataset.append(tokenized_data)
  return tokenized_dataset


def filter_dataset(
    tokenized_dataset: list[tuple[str, Any, str, int, int, int]],
    dataset_type: str,
    max_output_length: int = 0,
    run_mmlu_dataset: bool = False,
    min_input_length: int = 4,
    max_input_length: int = 0,
    max_target_length: int = 0,
    max_output_multiplier: int = 0,
) -> list[InputRequest]:
  print(
      f"In InputRequest, pass in max_output_length: {max_output_length} for"
      " each sample"
  )

  # Filter out too long sequences.
  filtered_dataset: list[InputRequest] = []
  for (
      prompt,
      _,
      output,
      prompt_len,
      output_len,
      sample_idx,
  ) in tokenized_dataset:
    if prompt_len < min_input_length or (
        not (run_mmlu_dataset or dataset_type == "math500") and output_len < 4
    ):
      # Prune too short sequences.
      # This is because TGI causes errors when the input or output length
      # is too short.
      # Math results could be really short though.
      continue
    if (
        prompt_len > max_input_length
        or prompt_len + output_len > max_target_length
    ):
      # Prune too long sequences.
      continue
    if dataset_type == "math500":
      max_output_len = max_output_length or output_len * max_output_multiplier
    else:
      max_output_len = max_output_length or output_len
    request = InputRequest(
        prompt, prompt_len, output, max_output_len, sample_idx
    )
    filtered_dataset.append(request)

  print(f"The dataset contains {len(tokenized_dataset)} samples.")
  print(f"The filtered dataset contains {len(filtered_dataset)} samples.")

  return filtered_dataset


def sample_requests(
    dataset: list[tuple[Any, Any]],
    tokenizer: Any,
    num_requests: int,
    dataset_type: str,
    max_output_length: int = 0,
    oversample_multiplier: float = 1.2,
    run_mmlu_dataset: bool = False,
    min_input_length: int = 4,
    max_input_length: int = 0,
    max_target_length: int = 0,
    max_output_multiplier: int = 0,
) -> list[InputRequest]:

  # Original dataset size
  n = len(dataset)
  dataset_indices = range(n)

  # Create necessary number of requests even if bigger than dataset size
  sampled_indices = random.sample(
      dataset_indices, min(int(num_requests * oversample_multiplier), n)
  )

  if num_requests > len(sampled_indices):
    print(
        f"Number of requests {num_requests} is larger than size of dataset"
        f" {n}.\n",
        "Repeating data to meet number of requests.\n",
    )
    sampled_indices = sampled_indices * int(
        np.ceil(num_requests / len(sampled_indices))
    )

  print(f"{len(sampled_indices)=}")
  # some of these will be filtered out, so sample more than we need

  sampled_dataset = []
  for i in sampled_indices:
    sampled_data = dataset[i] + (dataset_indices[i],)
    sampled_dataset.append(sampled_data)

  tokenized_dataset = tokenize_dataset(sampled_dataset, tokenizer)

  input_requests = filter_dataset(
      tokenized_dataset,
      dataset_type,
      max_output_length,
      run_mmlu_dataset,
      min_input_length,
      max_input_length,
      max_target_length,
      max_output_multiplier,
  )

  # Sample the requests.
  if len(input_requests) > num_requests:
    input_requests = random.sample(input_requests, num_requests)

  return input_requests


async def get_request(
    input_requests: list[InputRequest],
    request_rate: float,
) -> AsyncGenerator[InputRequest, None]:
  input_requests = iter(input_requests)
  for request in input_requests:
    yield request

    if request_rate == 0.0:
      # If the request rate is infinity, then we don't need to wait.
      continue
    # Sample the request interval from the exponential distribution.
    interval = np.random.exponential(1.0 / request_rate)
    # The next request will be sent after the interval.
    await asyncio.sleep(interval)


def calculate_metrics(
    input_requests: list[InputRequest],
    outputs: list[RequestFuncOutput],
    dur_s: float,
    tokenizer: Any,
) -> BenchmarkMetrics:
  total_output = 0
  total_input = 0
  completed = 0
  ttft = EventMetric("ttft", "Time-to-first-token", "ms")
  ttst = EventMetric("ttst", "Time-to-second-token", "ms")
  per_out_token_lat = EventMetric("TPOT", "Time-per-output-token", "ms")
  output_sizes = []
  for i in range(len(outputs)):
    if outputs[i].success:
      completed += 1
      output_len = len(
          outputs[i].generated_token_list
          if tokenizer != "test"
          else ["Ċ", "Ō", "Ɵ"]
      )
      output_sizes.append(output_len)
      total_output += output_len
      total_input += input_requests[i].prompt_len
      if output_len == 0:
        print(
            f"""-------- output_len is zero for {i}th request:,
             output: {outputs[i]}"""
        )
        continue
      ttft.record(outputs[i].ttft_sec * 1000)
      ttst.record(outputs[i].ttst_sec * 1000)
      per_out_token_lat.record(outputs[i].latency_sec / output_len * 1000)

  print("Mean output size:", float(np.mean(output_sizes)))
  print("Median output size:", float(np.median(output_sizes)))
  print("P99 output size:", float(np.percentile(output_sizes, 99)))

  metrics = BenchmarkMetrics(
      completed=completed,
      total_input=total_input,
      total_output=total_output,
      request_throughput=completed / dur_s,
      input_throughput=total_input / dur_s,
      output_throughput=total_output / dur_s,
      overall_throughput=(total_input + total_output) / dur_s,
      ttft=ttft,
      ttst=ttst,
      tpot=per_out_token_lat,
  )

  return metrics


async def grpc_async_request(
    api_url: str,
    request: Any,
    prefill_quota: AsyncCounter,
    active_req_quota: AsyncCounter,
    out_token_cnt: CounterMetric,
) -> tuple[list[int], float, float, float]:
  """Send grpc synchronous request since the current grpc server is sync."""
  options = [("grpc.keepalive_timeout_ms", 10000)]
  async with grpc.aio.insecure_channel(api_url, options=options) as channel:
    stub = jetstream_pb2_grpc.OrchestratorStub(channel)
    request_start_time = time.perf_counter()
    response = stub.Decode(request)
    token_list = []
    ttft = 0
    ttst = 0
    stream_resp_cnt = 0
    async for resp in response:
      stream_resp_cnt += 1
      if stream_resp_cnt == 1:
        await prefill_quota.inc()
        ttft = time.perf_counter() - request_start_time
        if ttft > 2.0:
          print(datetime.now(), f"slow TTFT {ttft:.2f}", prefill_quota.value())
      elif stream_resp_cnt == 2:
        ttst = time.perf_counter() - request_start_time
      resp_tokens = resp.stream_content.samples[0].token_ids
      token_list.extend(resp_tokens)
      out_token_cnt.increment(len(resp_tokens))
    await active_req_quota.inc()
    req_latency = time.perf_counter() - request_start_time
    return token_list, ttft, ttst, req_latency


async def send_request(
    api_url: str,
    tokenizer: Any,
    input_request: InputRequest,
    prefill_quota: AsyncCounter,
    active_req_quota: AsyncCounter,
    req_complete_cnt: CounterMetric,
    out_token_cnt: CounterMetric,
    pbar: tqdm,
) -> RequestFuncOutput:
  """Send the request to JetStream server."""
  # Tokenize on client side following MLPerf standard.
  token_ids = tokenizer.encode(input_request.prompt)

  # Send the request
  request = jetstream_pb2.DecodeRequest(
      token_content=jetstream_pb2.DecodeRequest.TokenContent(
          token_ids=token_ids
      ),
      max_tokens=input_request.output_len,
      metadata=jetstream_pb2.DecodeRequest.Metadata(
          start_time=time.perf_counter()
      ),
  )
  out_tokens, ttft_sec, ttst_sec, latency_sec = await grpc_async_request(
      api_url,
      request,
      prefill_quota,
      active_req_quota,
      out_token_cnt,
  )
  req_complete_cnt.increment()

  # Collect per-request output and metrics.
  output = RequestFuncOutput()
  output.input_request = input_request
  output.prompt_len = input_request.prompt_len
  output.ttft_sec = ttft_sec
  output.ttst_sec = ttst_sec
  output.latency_sec = latency_sec
  output.generated_token_list = out_tokens
  # generated_token_list is a list of token ids, decode it to generated_text.
  output.generated_text = tokenizer.decode(out_tokens)
  output.success = True
  if pbar:
    pbar.postfix = (
        f"#reqs: {active_req_quota.delta()}/"
        f"{active_req_quota.value()}; "
        f"#prefill: {prefill_quota.delta()}/"
        f"{prefill_quota.value()}"
    )
    pbar.update(1)
  return output


async def benchmark(
    api_url: str,
    tokenizer: Any,
    input_requests: list[InputRequest],
    request_rate: float,
    disable_tqdm: bool,
    prefill_quota: AsyncCounter,
    active_req_quota: AsyncCounter,
    is_warmup: bool = False,
) -> tuple[dict[str, float | int], list[RequestFuncOutput]]:
  """Benchmark the online serving performance.

  Args:
    api_url: URL (e.g. host:port) of the JetStream server to send requests to.
    tokenizer: The tokenizer used to convert texts into tokens that will be set
      in requests.
    input_requests: A list of requests to send.
    request_rate: The number of requests to send per second.
    disable_tqdm: Whether progress bar should be disabled or not.
    prefill_quota: Quota for limiting pending prefill operations.
    active_req_quota: Quota for limiting inflight requests.
    is_warmup: Whether this run is to warm up the server.

  Return:
    A tuple containing the performance statistics for all requests and a list
    of responses from the executed requests.
  """
  print(f"Benchmarking with a total number of {len(input_requests)} requests")
  print(f"Benchmarking with request rate of {request_rate}")
  pbar = None if disable_tqdm else tqdm(total=len(input_requests))
  req_complete_cnt = CounterMetric(
      "ReqCompleteCount", "Request Completion Counter"
  )
  out_token_cnt = CounterMetric("OutTokenCount", "OutToken Counter")

  # Run benchmarking
  tasks = []
  benchmark_start_time = time.perf_counter()
  async for request in get_request(input_requests, request_rate):
    await prefill_quota.dec()
    await active_req_quota.dec()
    tasks.append(
        asyncio.create_task(
            send_request(
                api_url=api_url,
                tokenizer=tokenizer,
                input_request=request,
                prefill_quota=prefill_quota,
                active_req_quota=active_req_quota,
                req_complete_cnt=req_complete_cnt,
                out_token_cnt=out_token_cnt,
                pbar=pbar,
            )
        )
    )
  outputs = await asyncio.gather(*tasks)
  if pbar is not None:
    pbar.close()

  # Compute metrics
  output_metrics = {}
  if not is_warmup:
    # No need to calculate metrics when executing warmup requests
    benchmark_duration = time.perf_counter() - benchmark_start_time
    metrics = calculate_metrics(
        input_requests=input_requests,
        outputs=outputs,
        dur_s=benchmark_duration,
        tokenizer=tokenizer,
    )
    print(f"Successful requests: {metrics.completed}")
    print(f"Benchmark duration: {benchmark_duration:2f} s")
    print(f"Total input tokens: {metrics.total_input}")
    print(f"Total generated tokens: {metrics.total_output}")
    print(f"Request throughput: {metrics.request_throughput:.2f} requests/s")
    print(f"Input token throughput: {metrics.input_throughput:.2f} tokens/s")
    print(f"Output token throughput: {metrics.output_throughput:.2f} tokens/s")
    print(
        f"Overall token throughput: {metrics.overall_throughput:.2f} tokens/s"
    )

    print(f"{metrics.ttft.distribution_summary_str()}")
    print(f"{metrics.ttst.distribution_summary_str()}")
    print(f"{metrics.tpot.distribution_summary_str()}")

    # Calculate one rate for each 10 sec window. Adjusts the window size if
    # needed to use csv output below for plotting the rate over time.
    window_size_sec = 10
    print(
        f"----- Request complete rate time series "
        f"(window_size = {window_size_sec} sec) -----"
    )
    print(f"{req_complete_cnt.rate_over_window_to_csv(window_size_sec)}")
    print(
        f"----- Output token rate time series "
        f"(window_size = {window_size_sec} sec) -----"
    )
    print(f"{out_token_cnt.rate_over_window_to_csv(window_size_sec)}")

    output_metrics = {
        "duration": benchmark_duration,
        "completed": metrics.completed,
        "total_input_tokens": metrics.total_input,
        "total_output_tokens": metrics.total_output,
        "request_throughput": metrics.request_throughput,
        "input_throughput": metrics.input_throughput,
        "output_throughput": metrics.output_throughput,
    }
    output_metrics = {
        **output_metrics,
        **metrics.ttft.distribution_summary_dict(),
        **metrics.ttst.distribution_summary_dict(),
        **metrics.tpot.distribution_summary_dict(),
    }
  return output_metrics, outputs


def mock_requests(total_mock_requests: int):
  """Generates a list of mock requests containing mock data."""
  data = []
  for _ in range(total_mock_requests):
    reqeust = InputRequest()
    reqeust.prompt = f"Prompt {random.randint(1, 1000)}"
    reqeust.prompt_len = random.randint(10, 100)
    reqeust.out = f"Output {random.randint(1, 1000)}"
    reqeust.output_len = random.randint(1, 10)
    data.append(reqeust)
  return data


def sample_warmup_requests(requests):
  interesting_buckets = [
      0,
      16,
      32,
      64,
      128,
      256,
      512,
      1024,
  ]

  for start, end in zip(interesting_buckets[:-1], interesting_buckets[1:]):
    for request in requests:
      if start < request.prompt_len <= end:
        yield request
        break


def parse_args() -> argparse.Namespace:
  parser = argparse.ArgumentParser(
      description="Benchmark the online serving throughput."
  )
  parser.add_argument(
      "--server",
      type=str,
      default="0.0.0.0",
      help="Server address.",
  )
  parser.add_argument("--port", type=str, default=9000)
  parser.add_argument(
      "--dataset",
      type=str,
      default="test",
      choices=[
          "test",
          "sharegpt",
          "openorca",
          "mmlu",
          "math500",
          "longcontext",
      ],
      help="The dataset name.",
  )
  parser.add_argument("--dataset-path", type=str, help="Path to the dataset.")
  parser.add_argument(
      "--model",
      type=str,
      default="no_model",
      help=(
          "Name of the model like llama-2, llama-3, gemma. (it's just used to"
          " label the benchmark, pick the tokenizer, the model config is"
          " defined in config_lib, and passed as the server config flag when"
          " we run the JetStream server)"
      ),
  )
  parser.add_argument(
      "--tokenizer",
      type=str,
      default="test",
      help=(
          "Name or path of the tokenizer. (For mock model testing, use the"
          " default value)"
      ),
  )
  parser.add_argument(
      "--use-hf-tokenizer",
      type=str2bool,
      default=False,
      help=(
          "Whether to use tokenizer from HuggingFace. If so, set this flag"
          " to True, and provide name of the tokenizer in the tokenizer flag."
      ),
  )
  parser.add_argument(
      "--num-prompts",
      type=int,
      default=1000,
      help=(
          "Number of prompts to process. (number of sample requests we randomly"
          " collect from dataset)"
      ),
  )
  parser.add_argument(
      "--request-rate",
      type=float,
      default=0.0,
      help=(
          "Number of requests per second. If this is 0., "
          "then all the requests are sent at time 0. "
          "Otherwise, we use Poisson process to synthesize "
          "the request arrival times."
      ),
  )
  parser.add_argument(
      "--total-mock-requests",
      type=int,
      default=150,
      help="The maximum number of mock requests to send for benchmark testing.",
  )
  parser.add_argument(
      "--max-output-length",
      type=int,
      default=0,
      help=(
          "The maximum output length for reference request. It would be passed"
          " to `max_tokens` parameter of the JetStream's DecodeRequest proto,"
          " and used in JetStream to control the output/decode length of a"
          " sequence. It would not be used in the engine. We should always set"
          " max_tokens <= (max_target_length - max_prefill_predict_length)."
          " max_target_length is the maximum length of a sequence;"
          " max_prefill_predict_length is the maximum length of the"
          " input/prefill of a sequence. Default to 0, in this case, "
          "the output length of the golden dataset would be passed."
      ),
  )
  parser.add_argument(
      "--max-target-length",
      type=int,
      default=2048,
      help=(
          "The maximum prompt length plus the output length for reference "
          " request. It would be used to filter the requests "
      ),
  )
  parser.add_argument(
      "--max-output-multiplier",
      type=int,
      default=2,
      help=(
          "The multiplier applied to the reference output length. The generated"
          " output might be longer than the reference outputs. We apply this "
          " multiplier for finer grained control. "
      ),
  )
  parser.add_argument(
      "--min-input-length",
      type=int,
      default=4,
      help=(
          "The minium input length for reference request. It would be used"
          " to filter the input requests. "
      ),
  )
  parser.add_argument(
      "--max-input-length",
      type=int,
      default=1024,
      help=(
          "The maximum input length for reference request. It would be used"
          " to filter the input requests. "
      ),
  )

  parser.add_argument("--seed", type=int, default=0)
  parser.add_argument(
      "--disable-tqdm",
      action="store_true",
      help="Specify to disable tqdm progress bar.",
  )
  parser.add_argument(
      "--save-result",
      action="store_true",
      help="Specify to save benchmark results to a json file",
  )
  parser.add_argument(
      "--additional-metadata-metrics-to-save",
      type=str,
      help=(
          "Additional metadata about the workload. Should be a dictionary in"
          " the form of a string."
      ),
  )
  parser.add_argument(
      "--save-request-outputs",
      action="store_true",
      help="Specify to store request outputs into a json file",
  )
  parser.add_argument(
      "--request-outputs-file-path",
      type=str,
      default="/tmp/request-outputs.json",
      help="File path to store request outputs",
  )
  parser.add_argument(
      "--run-eval",
      type=str2bool,
      default=False,
      help="Whether to run evaluation script on the saved outputs",
  )
  parser.add_argument(
      "--warmup-mode",
      type=str,
      default="none",
      choices=["none", "sampled", "full"],
      help="Whether to warmup first, and set the warmup mode",
  )
  parser.add_argument(
      "--conversation-starter",
      type=str,
      default="human",
      choices=["human", "gpt", "both"],
      help="What entity should be the one starting the conversations.",
  )
  parser.add_argument(
      "--num-shots",
      type=int,
      default=1,
      help="Num shorts to give for mmlu data set",
  )
  parser.add_argument(
      "--mmlu-method",
      type=str,
      default="HELM",
      choices=["HELM", "Harness", ""],
      help="mmlu method/format to generate shots",
  )
  parser.add_argument(
      "--run-mmlu-dataset",
      action="store_true",
      help="specify if it's for mmlu dataset",
  )
  return parser.parse_args()


def main(args: argparse.Namespace):
  print(args)
  random.seed(args.seed)
  np.random.seed(args.seed)

  model_id = args.model
  tokenizer_id = args.tokenizer
  use_hf_tokenizer = args.use_hf_tokenizer

  prefill_quota = AsyncCounter(init_value=3)
  active_req_quota = AsyncCounter(init_value=450)

  api_url = f"{args.server}:{args.port}"

  tokenizer = get_tokenizer(model_id, tokenizer_id, use_hf_tokenizer)
  if tokenizer == "test" or args.dataset == "test":
    input_requests = mock_requests(
        args.total_mock_requests
    )  # e.g. [("AB", 2, "AB", 3)]
  else:
    dataset = []
    if args.dataset == "openorca":
      dataset = load_openorca_dataset_pkl(args.dataset_path)
    elif args.dataset == "sharegpt":
      dataset = load_sharegpt_dataset(
          args.dataset_path,
          args.conversation_starter,
      )
    elif args.dataset == "mmlu":
      dataset = gen_mmlu_prompt(
          args.dataset_path, args.num_shots, args.mmlu_method
      )
    elif args.dataset == "math500":
      dataset = load_math500_dataset(
          args.dataset_path,
      )
    elif args.dataset == "longcontext":
      dataset = load_longcontext_dataset_pkl(
          args.dataset_path,
      )
    else:
      raise ValueError(
          f"Fatal Error: Uncognized input parameters: {args.dataset}"
      )

    # A given args.max_output_length value is the max generation step,
    # when the args.max_output_length is default to None, the sample's golden
    # output length will be used to decide the generation step.
    input_requests = sample_requests(
        dataset=dataset,
        tokenizer=tokenizer,
        num_requests=args.num_prompts,
        dataset_type=args.dataset,
        max_output_length=args.max_output_length,
        run_mmlu_dataset=args.run_mmlu_dataset,
        min_input_length=args.min_input_length,
        max_input_length=args.max_input_length,
        max_target_length=args.max_target_length,
        max_output_multiplier=args.max_output_multiplier,
    )

  warmup_requests = None
  if args.warmup_mode == "full":
    warmup_requests = input_requests
  elif args.warmup_mode == "sampled":
    warmup_requests = list(sample_warmup_requests(input_requests)) * 2

  if warmup_requests:
    print(f"Warmup (mode: {args.warmup_mode}) is starting.")
    _, _ = asyncio.run(
        benchmark(
            api_url=api_url,
            tokenizer=tokenizer,
            input_requests=warmup_requests,
            request_rate=args.request_rate,
            disable_tqdm=args.disable_tqdm,
            prefill_quota=prefill_quota,
            active_req_quota=active_req_quota,
            is_warmup=True,
        )
    )
    print(f"Warmup (mode: {args.warmup_mode}) has completed.")

  # TODO: Replace this with warmup complete signal once supported.
  # Wait for server completely warmup before running the benchmark.
  time.sleep(5)

  benchmark_result, request_outputs = asyncio.run(
      benchmark(
          api_url=api_url,
          tokenizer=tokenizer,
          input_requests=input_requests,
          request_rate=args.request_rate,
          disable_tqdm=args.disable_tqdm,
          prefill_quota=prefill_quota,
          active_req_quota=active_req_quota,
      )
  )

  # Process output
  output = [output.to_dict() for output in request_outputs]
  if args.run_eval:
    if args.run_mmlu_dataset:
      eval_json = eval_accuracy_mmlu(output)
    else:
      eval_json = eval_accuracy(output, args.dataset[:4])

  # Save config and results to json
  if args.save_result:
    # dimensions values are strings
    dimensions_json = {}
    # metrics values are numerical
    metrics_json = {}

    # Setup
    current_dt = datetime.now().strftime("%Y%m%d-%H%M%S")
    dimensions_json["date"] = current_dt
    dimensions_json["model_id"] = model_id
    dimensions_json["tokenizer_id"] = tokenizer_id
    if args.additional_metadata_metrics_to_save is not None:
      dimensions_json = {
          **dimensions_json,
          **json.loads(args.additional_metadata_metrics_to_save),
      }
    metrics_json["num_prompts"] = args.num_prompts

    # Traffic
    metrics_json["request_rate"] = args.request_rate
    metrics_json = {**metrics_json, **benchmark_result}
    if args.run_eval:
      metrics_json = {**metrics_json, **eval_json}

    final_json = {}
    final_json["metrics"] = metrics_json
    final_json["dimensions"] = dimensions_json

    # Save to file
    base_model_id = model_id.split("/")[-1]
    file_name = (
        f"JetStream-{args.request_rate}qps-{base_model_id}-{current_dt}.json"
    )
    with open(file_name, "w", encoding="utf-8") as outfile:
      json.dump(final_json, outfile)

  if args.save_request_outputs:
    file_path = args.request_outputs_file_path
    with open(file_path, "w", encoding="utf-8") as output_file:
      json.dump(
          output,
          output_file,
          indent=4,
      )


if __name__ == "__main__":
  gc.disable()
  main(parse_args())
